/**
  *  \file arch/unix/dir-arit.cc
  *  \brief Platform-specific directory arithmetic for Unix
  *
  *  This implements the platform-specific static Directory methods
  *  that deal with file names.
  */

#include <list>
#include "io/dirs.h"
#include "arch/platform.h"

/** Test whether file name is absolute.
    \returns true iff file name is absolute, that is, it has
    the same meaning on this computer no matter what the current
    working directory is. */
bool
Directory::isAbsoluteName(string_t name)
{
    return (name.length() && name[0] == Platform::PATH_SEPARATOR);
}

/** Test whether character is path separator.
    \returns true iff c is a path separator char (that is, something
    that can not occur in a file "basename") */
bool
Directory::isSeparatorChar(char c)
{
    return c == Platform::PATH_SEPARATOR;
}

/** Make file name from two components.
    \param dir     working directory
    \param name    file name
    \returns file name that names the file that would be accessed when
    we were in directory /dir/ and opened /name/. In particular, if
    /name/ is absolute, /dir/ is ignored. The returned file name is
    canonical. */
string_t
Directory::makeFileName(string_t dir, string_t name)
{
    if (isAbsoluteName(name))
        return getCanonicalName(name);
    if (dir.length() && !isSeparatorChar(dir[dir.length()-1]))
        dir += Platform::PATH_SEPARATOR;
    return getCanonicalName(dir + name);
}

/** Canonicalize directory name. Generates the canonical name for the
    specified file name/directory. It does so by removing redundant
    components (e.g. in Unix notation, `foo/./bar/../x' is the same
    as `foo/x').

    getCanonicalName("") returns the notation for the "current
    directory". */
string_t
Directory::getCanonicalName(string_t dir)
{
    if (!dir.length())
        return ".";

    bool is_absolute;
    int  up_count = 0;
    std::list<string_t> out;
    if (dir[0] == Platform::PATH_SEPARATOR) {
        is_absolute = true;
        dir.erase(0, 1);
    } else {
        is_absolute = false;
    }

    string_t::size_type pos = 0;
    while (pos < dir.length()) {
        string_t::size_type next = dir.find(Platform::PATH_SEPARATOR, pos);
        if (next == string_t::npos)
            next = dir.length();
        if (next == pos) {
            /* ignore consecutive slashes */
        } else if (next == pos+1 && dir[pos] == '.') {
            /* ignore "." components */
        } else if (next == pos+2 && dir.compare(pos, 2, "..", 2) == 0) {
            /* ".." component */
            if (!out.empty())
                out.pop_back();
            else
                ++up_count;
        } else {
            out.push_back(string_t(dir, pos, next-pos));
        }
        if (next >= dir.length())
            break;
        pos = next+1;
    }

    /* now generate output */
    string_t s;
    if (is_absolute) {
        s += Platform::PATH_SEPARATOR;
    } else {
        while (up_count--) {
            s += "..";
            s += Platform::PATH_SEPARATOR;
        }
    }
    for (std::list<string_t>::iterator i = out.begin(); i != out.end(); ++i) {
        s += *i;
        s += Platform::PATH_SEPARATOR;
    }
    if (s.length() > 1)
        s.erase(s.length() - 1);
    else if (! s.length())
        s = ".";                // these appear as result when input is "."
    return s;
}

/** Get file name part ("basename") of a name.
    This can be used to split a file name into pieces.
    \see getDirName
    \inv getAbsoluteName(dir) == getAbsoluteName(makeFileName(getDirName(dir), getFileName(dir))) */
string_t
Directory::getFileName(string_t dir)
{
    string_t::size_type d = dir.rfind(Platform::PATH_SEPARATOR);
    if (d != string_t::npos)
        return dir.substr(d+1);
    else
        return dir;
}

/** Get directory name part of a name. \see getFileName */
string_t
Directory::getDirName(string_t dir)
{
    string_t::size_type d = dir.rfind(Platform::PATH_SEPARATOR);
    if (d == 0)
        /* `/file' */
        return string_t(1, Platform::PATH_SEPARATOR);
    else if (d != string_t::npos)
        /* `dir/file' */
        return dir.substr(0, d);
    else
        /* `file' */
        return ".";
}
