/**
  *  \file client/chart/deletedrawingmode.cc
  *  \brief Starchart Mode: WDeleteDrawingChartMode
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chart/deletedrawingmode.h"
#include "client/chartloc.h"
#include "game/global.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "ui/event.h"
#include "ui/uiglobal.h"
#include "util/nls.h"

static void
deleteAdjacent(GDrawingContainer& c, GPoint pt)
{
    GDrawingContainer::iterator i = c.begin(), e = c.end();
    pt = pt.getSimpleCanonicalLocation();
    while (i != e) {
        GDrawing& d = **i;
        if (d.isVisible() && d.getType() == LineDrawing) {
            if (d.getPos().getSimpleCanonicalLocation() == pt) {
                // Starts at given position
                pt = d.getPos2().getSimpleCanonicalLocation();
                c.erase(i);
                i = c.begin();
            } else if (d.getPos2().getSimpleCanonicalLocation() == pt) {
                // Ends at given position
                pt = d.getPos().getSimpleCanonicalLocation();
                c.erase(i);
                i = c.begin();
            } else {
                // No match
                ++i;
            }
        } else {
            ++i;
        }
    }
}

WDeleteDrawingChartMode::WDeleteDrawingChartMode(GDrawingContainer::iterator p)
    : WMarkerChartMode(p), counter(0), phase(false)
{ }

WDeleteDrawingChartMode::~WDeleteDrawingChartMode()
{ }

void
WDeleteDrawingChartMode::drawCursor(GfxCanvas& can, GChartLocation& loc)
{
    if (*drawing) {
        uint8 color = (phase ? COLOR_WHITE : COLOR_RED);
        loc.getViewport().drawDrawing(can, **drawing, color, true /* FIXME: use user setting instead? */);
    }
    WMarkerChartMode::drawCursor(can, loc);
}

void
WDeleteDrawingChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    WMarkerChartMode::drawOverlays(can, loc);
    
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(COLOR_WHITE);

    GfxPoint sc = loc.getViewport().getCenter();
    string_t text = _(" Delete this drawing (y/n): ");
    int wi = ctx.getFont().getTextWidth(text);
    int x = sc.x - wi/2;
    int y = sc.y + 20;
    drawSolidBar(ctx, GfxRect(x, y, wi, font_heights[FONT_NORMAL]), COLOR_RED);
    outText(ctx, x, y, text);
    if (*drawing && (*drawing)->getType() == LineDrawing) {
        text = _(" Press [A] to delete adjacent lines, too. ");
        wi = ctx.getFont().getTextWidth(text);
        x = sc.x - wi/2;
        y += font_heights[FONT_NORMAL];
        drawSolidBar(ctx, GfxRect(x, y, wi, font_heights[FONT_NORMAL]), COLOR_RED);
        outText(ctx, x, y, text);
    }
}

bool
WDeleteDrawingChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    /* In PCC 1.x, Delete Mode eats all keyboard input except for
       (shifted) arrows, digits, and dot. Delete mode itself offers
       y (RET, SPC) and n (ESC), DEL to pick another object, and a
       to delete adjacent objects. */
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'y':
         case SDLK_RETURN:
         case ' ':
            // Delete marker
            if (*drawing) {
                GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
                c.erase(drawing);
            }
            delete this;
            return true;

         case 'n':
         case SDLK_ESCAPE:
         case vk_Quit:
            // Do not delete
            delete this;
            return true;

         case SDLK_DELETE:
            // Let top-level handle this!
            // FIXME: should normally be parent who handles this
            return false;

         case 'a':
            // Delete adjacent
            if (*drawing && (*drawing)->getType() == LineDrawing) {
                GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
                GPoint a = (*drawing)->getPos(), b = (*drawing)->getPos2();
                c.erase(drawing);
                deleteAdjacent(c, a);
                deleteAdjacent(c, b);
                delete this;
            }
            return true;

         default:
            {
                uint32 base_key = event.key.code & ~ss_AllMod;
                if (base_key == SDLK_LEFT || base_key == SDLK_RIGHT
                    || base_key == SDLK_UP || base_key == SDLK_DOWN
                    || base_key == vk_WheelUp || base_key == vk_WheelDown)
                {
                    // Handle these
                    return WMarkerChartMode::handleEvent(event, loc);
                } else {
                    // Swallow this key
                    return true;
                }
            }
        }
    }

    if (event.type == ev_Tick) {
        if (++counter >= TIMER_HZ/3) {
            counter = 0;
            phase = !phase;
            loc.markChanged();
        }
    }

    return WMarkerChartMode::handleEvent(event, loc);
}
