/**
  *  \file client/dialogs/gadialog.cc
  *  \brief Global Actions Dialog
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/gadialog.h"
#include "client/actions/gabuild.h"
#include "client/actions/gafcode.h"
#include "client/actions/gamission.h"
#include "client/actions/garemote.h"
#include "client/actions/gascript.h"
#include "client/actions/gatax.h"
#include "client/globalact.h"
#include "client/widgets/gaparamview.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/holder.h"
#include "cpluslib/ptrvector.h"
#include "game/config.h"
#include "game/host.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/widgets/button.h"
#include "ui/widgets/tree.h"
#include "util/nls.h"
#include "client/actions/gaexport.h"
#include "game/global.h"

namespace {
    class WGlobalActionDialog : public UIWindow {
        WGlobalActionParameter& parm;
        ObjectHolder holder;
        ptr_vector<WGlobalAction> actions;
        UITreeWidget* ptree;

        void prepare(UITreeWidget& tree);
        void init();
        void onExec();
     public:
        WGlobalActionDialog(WGlobalActionParameter& parm);
        ~WGlobalActionDialog();
    };
}

/********************************* Dialog ********************************/

/** Constructor.
    \param parm parameter object, must live longer than WGlobalActionDialog */
WGlobalActionDialog::WGlobalActionDialog(WGlobalActionParameter& parm)
    : UIWindow(_("Global Actions"), BLUE_WINDOW, COLOR_GRAY),
      parm(parm),
      ptree(0)
{
    init();
}

WGlobalActionDialog::~WGlobalActionDialog()
{ }

/** Prepare the tree of global actions. */
void
WGlobalActionDialog::prepare(UITreeWidget& tree)
{
    // Dummy element
    actions.push_back(0);

    // FCode actions
    tree.addItem(0, 0, _("Friendly Codes"));
    tree.addItem(actions.size(), 1, _("Randomize"));
    actions.push_back(new WGlobalRandomFCodeAction());
    tree.addItem(actions.size(), 1, _("Set to..."));
    actions.push_back(new WGlobalSetFCodeAction());

    // Remote Control actions
    if (host.isPHost() && config.CPEnableRemote()) {
        tree.addItem(0, 0, _("Remote Control"));
        tree.addItem(actions.size(), 1, _("Allow"));
        actions.push_back(new WGlobalRemoteControlAction(rcv_Allow));
        tree.addItem(actions.size(), 1, _("Forbid"));
        actions.push_back(new WGlobalRemoteControlAction(rcv_Forbid));
        tree.addItem(actions.size(), 1, _("Drop"));
        actions.push_back(new WGlobalRemoteControlAction(rcv_Drop));
        tree.addItem(actions.size(), 1, _("Request"));
        actions.push_back(new WGlobalRemoteControlAction(rcv_Control));
    }

    // Mission actions
    tree.addItem(0, 0, _("Missions"));
    tree.addItem(actions.size(), 1, _("Set starbase mission..."));
    actions.push_back(new WGlobalBaseMissionAction());
    tree.addItem(actions.size(), 1, _("Set starship mission..."));
    actions.push_back(new WGlobalShipMissionAction());
    tree.addItem(actions.size(), 1, _("Set starship primary enemy..."));
    actions.push_back(new WGlobalShipEnemyAction());

    // Build actions
    tree.addItem(0, 0, _("Building"));
    tree.addItem(actions.size(), 1, _("Set Build Goals..."));
    actions.push_back(new WGlobalSetBuildGoalsAction());
    tree.addItem(actions.size(), 1, _("Auto-build"));
    actions.push_back(new WGlobalAutobuildAction());

    // Taxation actions
    tree.addItem(0, 0, _("Taxes"));
    tree.addItem(actions.size(), 1, _("Optimize"));
    actions.push_back(new WGlobalAutotaxAction(WGlobalAutotaxAction::Both));
    tree.addItem(actions.size(), 1, _("Optimize (colonists only)"));
    actions.push_back(new WGlobalAutotaxAction(WGlobalAutotaxAction::Colonists));
    tree.addItem(actions.size(), 1, _("Optimize (natives only)"));
    actions.push_back(new WGlobalAutotaxAction(WGlobalAutotaxAction::Natives));

    // Export
    tree.addItem(actions.size(), 0, _("Export..."));
    actions.push_back(new WGlobalExportAction());

    // Script
    tree.addItem(actions.size(), 0, _("Script command..."));
    actions.push_back(new WGlobalScriptAction());

    // Finish
    tree.updateAfterModification();
}

/** Build dialog. */
void
WGlobalActionDialog::init()
{
    // VBox
    //   HBox
    //     VBox
    //       UITreeWidget
    //       UISpacer
    //     VBox
    //       WGlobalActionParameterView
    //       UISpacer
    //   HBox
    //     UIButton "Execute"
    //     UIButton "Close"
    //     UISpacer
    //     UIButton "Help"

    // FIXME: PCC 1.x displays helpful hints for all actions.
    // How can we do this?
    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1 = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g2 = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g11 = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g12 = holder.add(new UIGroup(UIVBoxLayout::instance5));

    UITreeWidget& tree = holder.add(new UITreeWidget(1000, 15, 20*font_ems[FONT_NORMAL]));
    ptree = &tree;
    prepare(tree);
    g11.add(tree);
    g11.add(holder.add(new UISpacer()));
    g12.add(holder.add(new WGlobalActionParameterView(parm)));
    g12.add(holder.add(new UISpacer()));
    g1.add(g11);
    g1.add(g12);

    g2.add(holder.add(new UIButton(0, _("Execute"), SDLK_RETURN)).addCall(this, &WGlobalActionDialog::onExec));
    g2.add(holder.add(new UIButton(1, _("Close"), SDLK_ESCAPE)).addStop());
    g2.add(holder.add(new UISpacer()));
    g2.add(holder.add(new UIButton(_("Help"), 'h')));

    add(g1);
    add(g2);
    add(holder.add(new WHelpWidget("pcc2:globact")));
    pack();
}

/** Execute selected action. */
void
WGlobalActionDialog::onExec()
{
    uint32_t index = ptree->getCurrentItemIndex();
    int32_t id = ptree->getItemId(index);
    if (id >= 0 && id < int32_t(actions.size()) && actions[id] != 0) {
        if (!parm.hasFlag(parm.OnlyShips | parm.OnlyPlanets)) {
            messageBox(_("Please select the \"Ships\" and/or \"Planets\" option before executing a global action."),
                       _("Global Actions"));
        } else {
            parm.execute(*actions[id]);
        }
    } else {
        if (ptree->hasChildren(index)) {
            ptree->toggleItem(index);
        }
    }
}

/** Global actions, search entry point.
    Perform global actions with a pre-filled object list.
    \param list object list */
void
doGlobalActions(GObjectList& list)
{
    WGlobalActionParameter parm(list);
    WGlobalActionDialog dlg(parm);

    ui_root->centerWidget(dlg);
    ui_root->runModal(dlg);
}

/** Global actions, regular entry point.
    Perform global actions called as a standalone function. */
void
doGlobalActions()
{
    GObjectList list;
    doGlobalActions(list);
}
