/**
  *  \file client/dialogs/imperial.cc
  *  \brief Imperial Statistics
  *
  *  This is an almost straight-forward port of PCC 1.x's imperial.pas,
  *  producing 99% identical reports in much the same way.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/dialogs/imperial.h"
#include "client/widgets/decaymsg.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "cpluslib/misc.h"
#include "game/config.h"
#include "game/cspec.h"
#include "game/global.h"
#include "game/host.h"
#include "game/planet.h"
#include "game/pref.h"
#include "game/racename.h"
#include "game/ship.h"
#include "game/tables.h"
#include "game/team.h"
#include "gfx/palette.h"
#include "int/if/userif.h"
#include "int/simple.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/richdocument.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/fixedspacer.h"
#include "ui/window.h"
#include "util/misc.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/strlist.h"
#include "game/planetform.h"
#include "client/dialogbinding.h"

namespace {
    /* Pages */
    enum {
        TotalsPage,
        MineralsPage,
        PlanetsPage,
        ColonyPage,
        StarbasePage,
        StarshipPage,
        CapitalPage,
        StarchartPage,
        WeaponsPage,
        LastPage
    };

    /* Option flags for all pages. Names and values taken from PCC 1.x. */
    enum {
        // Sort orders for S/C/B
        sort_by_amount  = 0,
        sort_by_name    = 1,
        sort_by_id      = 2,
        sort_by_mass    = 3,
        sort_by_tech    = 4,
        sort_by_cargo   = 5,
        sort_by_engines = 6,
        sort_by_beams   = 7,
        sort_by_tubes   = 8,
        sort_by_bays    = 9,
        sort_mask       = 0x0F,
        hide_top        = 0x10,

        // Filters for O
        show_only_Colonists = 16,
        show_only_Supplies  = 32,
        show_only_Money     = 48,

        // Sort orders for M
        sort_by_total  = 0,
        sort_by_mined  = 1,
        show_only_N    = 16,
        show_only_T    = 32,
        show_only_D    = 48,
        show_only_M    = 64,

        // Filters for W
        show_everything = 0,
        show_only_Beams = 1,
        show_only_Torps = 2,
        show_only_Rest  = 3,

        // Sort orders for P
        sort_P_by_NumPlanets = 0,
        sort_P_by_TotalPop   = 1,
        sort_P_by_Race       = 2
    };

    /* Options */
    static uint8_t pageOptions[LastPage];
    static uint8_t lastActivePage;
}

/** Sort key array. Shuffles the keyArray so that dataArray[keyArray[i]] is a
    decreasing sequence.
    \param keyArray  List of keys, nkeys values, initialized [0,nkeys)
    \param dataArray Data values, nkeys values
    \param nkeys     Number of items */
static void
sortKeys(int keyArray[], int32_t dataArray[], uint32_t nkeys)
{
    uint32_t i = 1;
    while (i < nkeys) {
        if (dataArray[keyArray[i-1]] < dataArray[keyArray[i]]) {
            std::swap(keyArray[i-1], keyArray[i]);
            if (i > 1)
                --i;
        } else {
            ++i;
        }
    }
}

/** Fill key array. Initializes a keyArray for sortKeys.
    \param keyArray [out] List of keys, nkeys values, initialized [0,nkeys)
    \param nkeys    Number of items */
static void
fillKeys(int keyArray[], uint32_t nkeys)
{
    for (uint32_t i = 0; i < nkeys; ++i)
        keyArray[i] = i;
}

/****************************** Totals Page ******************************/

static const GCargoType ctypes[] = {
    el_Neutronium,
    el_Tritanium,
    el_Duranium,
    el_Molybdenum,
    el_Supplies,
    el_Colonists,
    el_Money
};

/** Count minerals from all our planets. */
static void
sumMinedMinerals(GCargoSpec& min, GPlayedPlanetType& type)
{
    for (int i = type.findNextIndex(0); i != 0; i = type.findNextIndex(i)) {
        const GPlanet& pl = type.getObjectByIndex(i);
        for (std::size_t j = 0; j < countof(ctypes); ++j) {
            min.add(ctypes[j], pl.getCargo(ctypes[j]));
        }
    }
}

/** Add a ship transfer's cargo to our total sum.
    \param min [in/out] Accumulator
    \param sh  [in] Ship
    \param tp  [in] Which transporter */
static void
sumShipTransfer(GCargoSpec& min, const GShip& sh, const GShip::Transporter tp)
{
    if (sh.isTransporterActive(tp)) {
        const TShipTransfer& t = sh.getTransporter(tp);
        min.add(el_Neutronium, t.ore[el_Neutronium]);
        min.add(el_Tritanium, t.ore[el_Tritanium]);
        min.add(el_Duranium, t.ore[el_Duranium]);
        min.add(el_Molybdenum, t.ore[el_Molybdenum]);
        min.add(el_Colonists, t.colonists);
        min.add(el_Supplies, t.supplies);
    }
}

/** Count minerals from all our ships. */
static void
sumShipMinerals(GCargoSpec& min, GPlayedShipType& type)
{
    for (int i = type.findNextIndex(0); i != 0; i = type.findNextIndex(i)) {
        const GShip& sh = type.getObjectByIndex(i);
        for (std::size_t j = 0; j < countof(ctypes); ++j) {
            min.add(ctypes[j], sh.getCargo(ctypes[j]));
        }
        sumShipTransfer(min, sh, GShip::t_Unload);
        sumShipTransfer(min, sh, GShip::t_Transfer);
    }
}

/** Count ground minerals, production, and max production. */
static void
sumGroundMineralsAndProduction(GCargoSpec& ground,
                               GCargoSpec& produced,
                               GCargoSpec& maxProduced,
                               GPlayedPlanetType& type)
{
    for (int i = type.findNextIndex(0); i != 0; i = type.findNextIndex(i)) {
        const GPlanet& pl = type.getObjectByIndex(i);
        for (int i = el_Neutronium; i <= el_Molybdenum; ++i) {
            int32_t gnd = pl.getOreGround(i);
            ground.add(GCargoType(i), gnd);

            int capacity = getMiningCapacity(pl, i, pl.getStructures(str_Mines));
            maxProduced.add(GCargoType(i), capacity);

            if (capacity > gnd)
                capacity = gnd;
            produced.add(GCargoType(i), capacity);
        }

        /* Bovinoids */
        produced.add(el_Supplies, getBovinoidSupplyContributionLimited(pl));
        maxProduced.add(el_Supplies, getBovinoidSupplyContribution(pl));

        /* Factories */
        int32_t fact = pl.getStructures(str_Factories);
        if (host.isPHost())
            fact = fact * config.ProductionRate(pl.getOwner()) / 100;
        produced.add(el_Supplies, fact);
        maxProduced.add(el_Supplies, fact);

        /* Taxes */
        int32_t due = getColonistDue(pl, pl.getColonistTax());
        int32_t rem;
        int32_t pay = getColonistDueLimited(pl, pl.getColonistTax(), rem);
        if (pl.getNatives() > 0) {
            due += getNativeDue(pl, pl.getNativeTax());
            pay += getNativeDueLimited(pl, pl.getNativeTax(), rem);
        }
        produced.add(el_Money, pay);
        maxProduced.add(el_Money, due);
    }
}

/** Count structures.
    \param structures [in/out] Accumulator; indexed by str_XXX */
static void
sumStructures(int structures[], GPlayedPlanetType& type)
{
    for (int i = type.findNextIndex(0); i != 0; i = type.findNextIndex(i)) {
        const GPlanet& pl = type.getObjectByIndex(i);
        for (int s = 0; s < 3; ++s)
            structures[s] += pl.getStructures(TPlanetaryStructure(s));
    }
}

/** Show a line containing two cargo values.
    \param d  [in/out] document to write to
    \param el [in] Cargo type to use
    \param a  [in] Left column value
    \param b  [in] Right column value */
static void
addCargoPair(RichDocument& d, GCargoType el, const GCargoSpec& a, const GCargoSpec& b)
{
    const int em = font_ems[FONT_NORMAL];
    d.add(string_t(getCargoName(el) + ":"));
    d.addRight(15*em, RichText(UIColor::tc_Green, numToString(a.get(el))));
    d.add(RichText(UIColor::tc_Green, " " + getCargoUnit(el)));
    d.addRight(25*em, RichText(UIColor::tc_Green, numToString(b.get(el))));
    d.add(RichText(UIColor::tc_Green, " " + getCargoUnit(el)));
    d.addNewline();
}

/** Draw "Totals" page. */
static void
drawTotalsPage(RichDocument& d, GUniverse& univ)
{
    typedef RichTextStyleAttribute RSA;
    const int em = font_ems[FONT_NORMAL];

    d.add(RichText(UIColor::tc_White, _("Totals")).withAttribute(RSA::Big));
    d.addParagraph();

    d.add(_("Planets:"));
    d.addRight(15*em, RichText(UIColor::tc_Green, numToString(univ.ty_played_planets.countObjects())));
    d.addNewline();
    d.add(_("Starbases:"));
    d.addRight(15*em, RichText(UIColor::tc_Green, numToString(univ.ty_played_bases.countObjects())));
    d.addNewline();

    int numCapital = 0;
    for (int sid = univ.ty_played_ships.findNextIndex(0); sid != 0; sid = univ.ty_played_ships.findNextIndex(sid)) {
        const GShip& sh = univ.ty_played_ships.getObjectByIndex(sid);
        if (sh.getNumBays() > 0 || sh.getNumBeams() > 0 || sh.getNumTorpLaunchers() > 0)
            ++numCapital;
    }
    d.add(_("Starships:"));
    d.addRight(15*em, RichText(UIColor::tc_Green, numToString(univ.ty_played_ships.countObjects())));
    d.addNewline();
    d.add(_("  Capital ships:"));
    d.addRight(15*em, RichText(UIColor::tc_Green, numToString(numCapital)));
    d.addParagraph();

    d.addAt(0, RichText(UIColor::tc_White, _("Minerals")));
    d.addRight(15*em, _("(available)"));
    d.addRight(25*em, _("(ground)"));
    d.addNewline();

    GCargoSpec availableResources, groundResources, producedResources, maxProducedResources;
    int32_t structures[4];
    for (int i = 0; i < 4; ++i)
        structures[i] = 0;
    sumMinedMinerals(availableResources, univ.ty_played_planets);
    sumShipMinerals(availableResources, univ.ty_played_ships);
    sumGroundMineralsAndProduction(groundResources, producedResources, maxProducedResources, univ.ty_played_planets);
    sumStructures(structures, univ.ty_played_planets);

    for (int i = el_Neutronium; i <= el_Molybdenum; ++i) {
        addCargoPair(d, GCargoType(i), availableResources, groundResources);
    }
    d.addNewline();

    d.add(RichText(UIColor::tc_White, _("Colonies")));
    d.addNewline();
    d.add(_("Colonists:"));
    d.addRight(15*em, RichText(UIColor::tc_Green, clansToString(availableResources.get(el_Colonists))));
    d.addNewline();
    d.add(_("Money:"));
    d.addRight(15*em, RichText(UIColor::tc_Green, numToString(availableResources.get(el_Money))));
    d.add(RichText(UIColor::tc_Green, _(" mc")));
    d.addNewline();
    d.add(_("Supplies:"));
    d.addRight(15*em, RichText(UIColor::tc_Green, numToString(availableResources.get(el_Supplies))));
    d.add(RichText(UIColor::tc_Green, _(" kt")));
    d.addNewline();
    d.add(_("Mineral Mines:"));
    d.addRight(15*em, RichText(UIColor::tc_Green, numToString(structures[str_Mines])));
    d.addNewline();
    d.add(_("Factories:"));
    d.addRight(15*em, RichText(UIColor::tc_Green, numToString(structures[str_Factories])));
    d.addNewline();
    d.add(_("Defense Posts:"));
    d.addRight(15*em, RichText(UIColor::tc_Green, numToString(structures[str_Defense])));
    d.addParagraph();

    d.add(RichText(UIColor::tc_White, _("Production")));
    d.addRight(25*em, _("(max)"));
    d.addNewline();
    for (int i = el_Neutronium; i <= el_Molybdenum; ++i) {
        addCargoPair(d, GCargoType(i), producedResources, maxProducedResources);
    }
    addCargoPair(d, el_Money, producedResources, maxProducedResources);
    addCargoPair(d, el_Supplies, producedResources, maxProducedResources);
}

/***************************** Minerals Page *****************************/

/** Show top mineral planets.
    \param d     [in/out] Document to write to
    \param opts  [in] Page options
    \param count [in] Number of planets to write
    \param el    [in] Element to write */
static void
showTopMinerals(RichDocument& d, GUniverse& univ, const int opts, const int count, const GCargoType el)
{
    const int em = font_ems[FONT_NORMAL];

    d.add(RichText(UIColor::tc_White, format(_("Top %d %s Planets"), count, getCargoName(el))));
    d.addRight(22*em, _("(total)"));
    d.addRight(32*em, _("(mined)"));
    d.addNewline();

    /* Acquire data (that's 6 kbyte, I hope that fits on today's stack :-) */
    int ids[NUM_PLANETS+1];
    int32_t total[NUM_PLANETS+1];
    int32_t mined[NUM_PLANETS+1];
    fillKeys(ids, NUM_PLANETS+1);
    std::fill_n(total, countof(total), 0);
    std::fill_n(mined, countof(mined), 0);

    GPlayedPlanetType& type = univ.ty_played_planets;
    for (int i = type.findNextIndex(0); i != 0; i = type.findNextIndex(i)) {
        const GPlanet& pl = type.getObjectByIndex(i);
        mined[i] = pl.getCargo(el);
        total[i] = mined[i] + pl.getOreGround(el);
    }

    /* Sort */
    if ((opts & 0x0F) == sort_by_mined) {
        sortKeys(ids, mined, NUM_PLANETS+1);
    } else {
        sortKeys(ids, total, NUM_PLANETS+1);
    }

    /* Draw */
    for (int i = 0; i < count; ++i) {
        int pid = ids[i];
        if (total[pid] > 0) {
            d.add(RichText(type.getObjectByIndex(pid).getName(GObject::LongName))
                  .withAttribute(new RichTextLinkAttribute(format("q:UI.GotoScreen 2,%d", pid))));
            d.addRight(22*em, RichText(UIColor::tc_Green, numToString(total[pid])));
            d.add(RichText(UIColor::tc_Green, _(" kt")));
            d.addRight(32*em, RichText(UIColor::tc_Green, numToString(mined[pid])));
            d.add(RichText(UIColor::tc_Green, _(" kt")));
            d.addNewline();
        }
    }
    d.addNewline();
}

static void
drawMineralsPage(RichDocument& d, GUniverse& univ, int opts)
{
    d.add(RichText(UIColor::tc_White, _("Minerals")).withAttribute(RichTextStyleAttribute::Big));
    d.addParagraph();

    switch (opts & 0xF0) {
     case show_only_N:
        showTopMinerals(d, univ, opts, 24, el_Neutronium);
        break;
     case show_only_T:
        showTopMinerals(d, univ, opts, 24, el_Tritanium);
        break;
     case show_only_D:
        showTopMinerals(d, univ, opts, 24, el_Duranium);
        break;
     case show_only_M:
        showTopMinerals(d, univ, opts, 24, el_Molybdenum);
        break;
     default:
        showTopMinerals(d, univ, opts, 5, el_Neutronium);
        showTopMinerals(d, univ, opts, 5, el_Tritanium);
        showTopMinerals(d, univ, opts, 5, el_Duranium);
        showTopMinerals(d, univ, opts, 5, el_Molybdenum);
        break;
    }
}

/****************************** Planets Page *****************************/

static void
drawPlanetsPage(RichDocument& d, GUniverse& univ, int opts)
{
    typedef RichTextStyleAttribute RSA;
    const int em = font_ems[FONT_NORMAL];
    const UIColor::Color gr = UIColor::tc_Green;
    int dfu = config.DefenseForUndetectable();

    /* Data */
    enum { MAX = nr_Siliconoid };
    int32_t nativePlanets[MAX+1];
    int32_t nativePop[MAX+1];
    int nativeIndex[MAX+1];
    std::fill_n(nativePlanets, MAX+1, 0);
    std::fill_n(nativePop,     MAX+1, 0);
    int32_t nUndefended = 0;
    int32_t nVisible = 0;
    int32_t nCount = 0;

    /* Compute */
    GPlayedPlanetType& type = univ.ty_played_planets;
    for (int pid = type.findNextIndex(0); pid != 0; pid = type.findNextIndex(pid)) {
        const GPlanet& pl = type.getObjectByIndex(pid);
        if (pl.getNativeRace() >= 0 && pl.getNativeRace() <= MAX) {
            ++nativePlanets[pl.getNativeRace()];
            nativePop[pl.getNativeRace()] += pl.getNatives();
        }
        ++nCount;
        if (pl.getStructures(str_Defense) < dfu)
            ++nVisible;
        if (pl.getStructures(str_Defense) < 10)
            ++nUndefended;
    }

    /* Sort */
    fillKeys(nativeIndex, MAX+1);
    switch (opts) {
     case sort_P_by_NumPlanets:
        sortKeys(nativeIndex, nativePlanets, MAX+1);
        break;
     case sort_P_by_TotalPop:
        sortKeys(nativeIndex, nativePop, MAX+1);
        break;
     case sort_P_by_Race:
     default:
        break;
    }

    /* Draw */
    d.add(RichText(UIColor::tc_White, _("Planets")).withAttribute(RSA::Big));
    d.addParagraph();
    d.add(_("Total:"));
    d.addRight(15*em, RichText(gr, numToString(nCount)));
    d.addParagraph();
    if (nCount == 0)
        return;

    d.addAt(0, RichText(UIColor::tc_White, _("Natives")));
    d.addRight(15*em, _("Planets"));
    d.addRight(25*em, _("Natives"));
    d.addNewline();
    for (int i = 0; i <= MAX; ++i) {
        int race = nativeIndex[i];
        if (opts == sort_P_by_Race || nativePlanets[race] != 0) {
            RichText str(getNativeRaceName(race));
            if (nativePlanets[race] != 0)
                str.withAttribute(new RichTextLinkAttribute(format("q:UI.Search 'Owner$=My.Race$ And Natives.Race$=%d', 'p2'", race)));
            d.add(str);
            d.addRight(15*em, RichText(gr, numToString(nativePlanets[race])));
            if (race != 0 && nativePlanets[race])
                d.addRight(25*em, RichText(gr, clansToString(nativePop[race])));
            d.addNewline();
        }
    }
    d.addNewline();

    {
        RichText str(_("Nearly undefended:"));
        if (nUndefended > 0)
            str.withAttribute(new RichTextLinkAttribute("q:UI.Search 'Defense<10 And Owner$=My.Race$', 'p2'"));
        d.add(str);
        d.addRight(15*em, RichText(gr, numToString(nUndefended)));
        d.addNewline();
    }
    {
        RichText str(_("Visible by sensor scan:"));
        if (nUndefended > 0)
            str.withAttribute(new RichTextLinkAttribute(format("q:UI.Search 'Defense<%d And Owner$=My.Race$', 'p2'", dfu)));
        d.add(str);
        d.addRight(15*em, RichText(gr, numToString(nVisible)));
        d.addNewline();
    }    
}

/******************************* ColonyPage ******************************/

/** Show top resource planets.
    \param d     [in/out] Document to write to
    \param opts  [in] Page options
    \param count [in] Number of planets to write
    \param el    [in] Element to write */
static void
showTopResources(RichDocument& d, GUniverse& univ, const int opts, const int count, const GCargoType el)
{
    (void) opts;
    const int em = font_ems[FONT_NORMAL];

    d.add(RichText(UIColor::tc_White, format(_("Top %d %s Planets"), count, getCargoName(el))));
    d.addRight(22*em, string_t(format("(%s)", getCargoUnit(el))));
    d.addNewline();

    /* Acquire data */
    int ids[NUM_PLANETS+1];
    int32_t amount[NUM_PLANETS+1];
    fillKeys(ids, NUM_PLANETS+1);
    std::fill_n(amount, countof(amount), 0);

    GPlayedPlanetType& type = univ.ty_played_planets;
    for (int i = type.findNextIndex(0); i != 0; i = type.findNextIndex(i)) {
        amount[i] = type.getObjectByIndex(i).getCargo(el);
    }

    /* Sort */
    sortKeys(ids, amount, NUM_PLANETS+1);

    /* Draw */
    for (int i = 0; i < count; ++i) {
        int pid = ids[i];
        if (amount[pid] > 0) {
            d.add(RichText(type.getObjectByIndex(pid).getName(GObject::LongName))
                  .withAttribute(new RichTextLinkAttribute(format("q:UI.GotoScreen 2,%d", pid))));
            d.addRight(22*em, RichText(UIColor::tc_Green, numToString(amount[pid])));
            d.addNewline();
        }
    }
    d.addNewline();
}

static void
drawColonyPage(RichDocument& d, GUniverse& univ, int opts)
{
    d.add(RichText(UIColor::tc_White, _("Minerals")).withAttribute(RichTextStyleAttribute::Big));
    d.addParagraph();

    switch (opts & 0xF0) {
     case show_only_Colonists:
        showTopResources(d, univ, opts, 24, el_Colonists);
        break;
     case show_only_Supplies:
        showTopResources(d, univ, opts, 24, el_Supplies);
        break;
     case show_only_Money:
        showTopResources(d, univ, opts, 24, el_Money);
        break;
     default:
        showTopResources(d, univ, opts, 5, el_Colonists);
        showTopResources(d, univ, opts, 5, el_Supplies);
        showTopResources(d, univ, opts, 5, el_Money);
        break;
    }
}

/***************************** Starbase Page *****************************/

static bool
compareHulls(const int order, const std::vector<int>& counts, const int a, const int b)
{
    switch (order) {
     case sort_by_amount:
        return counts[a] > counts[b];
     case sort_by_name:
        return getHull(a).getName() < getHull(b).getName();
     case sort_by_id:
        return a < b;
     case sort_by_mass:
        return getHull(a).getMass() > getHull(b).getMass();
     case sort_by_tech:
        return getHull(a).getTech() > getHull(b).getTech();
     case sort_by_cargo:
        return getHull(a).getCargo() > getHull(b).getCargo();
     case sort_by_engines:
        return getHull(a).getNumEngines() > getHull(b).getNumEngines();
     case sort_by_beams:
        return getHull(a).getMaxBeams() > getHull(b).getMaxBeams();
     case sort_by_tubes:
        return getHull(a).getMaxLaunchers() > getHull(b).getMaxLaunchers();
     case sort_by_bays:
        return getHull(a).getNumBays() > getHull(b).getNumBays();
     default:
        return a < b;
    }
}

static void
drawHullList(RichDocument& d, std::vector<int>& counts, const int order, const char* query_fmt)
{
    /* Same stupid algorithm as in PCC 1.x - selection sort.
       We cannot easily use sortKeys here because we allow sorting by a string */
    const int em = font_ems[FONT_NORMAL];
    while (1) {
        int hid = 0;
        for (std::size_t i = 1; i < counts.size(); ++i) {
            if (counts[i] > 0) {
                if (hid == 0 || compareHulls(order, counts, i, hid)) {
                    hid = i;
                }
            }
        }
        if (hid != 0) {
            d.add(RichText(getHull(hid).getName())
                  .withAttribute(new RichTextLinkAttribute(format(query_fmt, hid))));
            d.addRight(22*em, RichText(UIColor::tc_Green, numToString(counts[hid])));
            d.addNewline();
            counts[hid] = 0;
        } else {
            break;
        }
    }
}

static void
drawStarbasePage(RichDocument& d, GUniverse& univ, int opts)
{
    enum { ITEMS = 8 };
    static const char*const names[ITEMS] = {
        N_("Tech 10 Hulls"),
        N_("Tech 10 Engines"),
        N_("Tech 10 Beams"),
        N_("Tech 10 Torpedoes"),
        N_("Building a ship"),
        N_("Recycling a ship"),
        N_("Repairing a ship"),
        N_("Have parts in storage"),
    };
    static const char*const expr[ITEMS] = {
        "Tech.Hull=10",
        "Tech.Engine=10",
        "Tech.Beam=10",
        "Tech.Torpedo=10",
        "Build",
        "Shipyard.Action=\"Recycle\"",
        "Shipyard.Action=\"Fix\"",
        "Storage.Hulls(0)+Storage.Engines(0)+Storage.Beams(0)+Storage.Launchers(0)"
    };

    /* Data */
    std::vector<int32_t> builds(getNumHulls()+1);
    int counts[ITEMS];
    int nbases = 0;
    std::fill_n(counts, std::size_t(ITEMS), 0);

    GPlayedBaseType& type = univ.ty_played_bases;
    for (int pid = type.findNextIndex(0); pid != 0; pid = type.findNextIndex(pid)) {
        const GPlanet& pl = type.getObjectByIndex(pid);
        ++nbases;
        if (pl.getBaseTechLevel(tl_Hull) == 10)
            ++counts[0];
        if (pl.getBaseTechLevel(tl_Engine) == 10)
            ++counts[1];
        if (pl.getBaseTechLevel(tl_Beam) == 10)
            ++counts[2];
        if (pl.getBaseTechLevel(tl_Torpedo) == 10)
            ++counts[3];
        if (pl.getBaseBuildHull() > 0 && pl.getBaseBuildHull() <= getNumHulls()) {
            ++counts[4];
            ++builds[pl.getBaseBuildHull()];
        } else {
            ++builds[0];
        }
        for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i) {
            if ((i <= NUM_ENGINES && pl.getBaseEngineStore(i) > 0)
                || (i <= NUM_BEAMS && pl.getBaseBeamStore(i) > 0)
                || (i <= NUM_TORPS && pl.getBaseLauncherStore(i) > 0)
                || pl.getBaseHullStoreSlot(i) > 0)
            {
                ++counts[7];
                break;
            }
        }
        if (pl.getBaseShipyardOrder() == sya_Recycle)
            ++counts[5];
        if (pl.getBaseShipyardOrder() == sya_Fix)
            ++counts[6];
    }

    /* Draw */
    const int em = font_ems[FONT_NORMAL];
    d.add(RichText(UIColor::tc_White, _("Starbases")).withAttribute(RichTextStyleAttribute::Big));
    d.addParagraph();
    if ((opts & hide_top) == 0) {
        d.add(_("Total:"));
        d.addRight(20*em, RichText(UIColor::tc_Green, numToString(nbases)));
        if (nbases == 0)
            return;
        d.addNewline();

        for (int i = 0; i < ITEMS; ++i) {
            RichText str((string_t(_(names[i])) + ":"));
            if (counts[i] > 0)
                str.withAttribute(new RichTextLinkAttribute(format("q:UI.Search '%s', '2b'", expr[i])));
            d.add(str);
            d.addRight(20*em, RichText(UIColor::tc_Green, numToString(counts[i])));
            d.addNewline();
        }
    }
    d.addNewline();

    d.add(RichText(UIColor::tc_White, _("Ships Being Built")));
    d.addNewline();

    drawHullList(d, builds, opts & sort_mask, "q:UI.Search 'Build.Hull$=%d', '2b'");

    if (counts[4] == 0) {
        d.add(_("(none)"));
        d.addNewline();
    }
}

/***************************** Starship Page *****************************/

/** Draw starship page.
    \param d [in/out] Document to write to
    \param opts [in] Page options
    \param withFreighters [in] true to include freighters, false for capital ships only */
static void
drawStarshipPage(RichDocument& d, GUniverse& univ, const int opts, const bool withFreighters)
{
    static const int ITEMS = 8;
    static const char*const names[] = {
        N_("In free space:"),
        N_("Carriers:"),
        N_("Torpedo Ships:"),
        N_("Ships w/o fuel:"),
        N_("Damaged Ships:"),
        N_("Gravitonic accel:"),
        N_("Alchemy Ships:"),
        N_("Cloakable Ships:"),
    };
    static const char*const exprs[] = {
        "Orbit$=0",
        "Type.Short=\"C\"",
        "Type.Short=\"T\"",
        "Cargo.N=0",
        "Damage<>0",
        "InStr(Hull.Special,\"G\")",
        "InStr(Hull.Special,\"A\")",
        "InStr(Hull.Special,\"C\")",
    };

    /* Acquire data */
    std::vector<int32_t> hullCounts(getNumHulls()+1);
    int32_t levelCounts[NUM_EXPERIENCE_LEVELS+1];
    int32_t counts[ITEMS];
    int32_t count = 0;
    int32_t levels = 0;
    std::fill_n(levelCounts, countof(levelCounts), 0);
    std::fill_n(counts, countof(counts), 0);

    GPlayedShipType& type = univ.ty_played_ships;
    for (int sid = type.findNextIndex(0); sid != 0; sid = type.findNextIndex(sid)) {
        const GShip& sh = type.getObjectByIndex(sid);
        if (withFreighters || sh.getNumBeams() > 0 || sh.getNumBays() > 0 || sh.getNumTorpLaunchers() > 0) {
            ++count;
            if (sh.getNumTorpLaunchers() > 0)
                ++counts[2];
            if (sh.getNumBays() > 0)
                ++counts[1];
            ++hullCounts[sh.getHullId()];
            if (sh.getCargo(el_Neutronium) == 0)
                ++counts[3];
            if (sh.getDamage() > 0)
                ++counts[4];
            if (sh.canDoSpecial(hf_Cloak) || sh.canDoSpecial(hf_AdvancedCloak) || sh.canDoSpecial(hf_HardenedCloak))
                ++counts[7];
            if (sh.canDoSpecial(hf_Gravitonic))
                ++counts[5];
            if (sh.canDoSpecial(hf_MerlinAlchemy) || sh.canDoSpecial(hf_AriesRefinery) || sh.canDoSpecial(hf_NeutronicRefinery))
                ++counts[6];
            if (sh.getOrbitPlanetId() == 0)
                ++counts[0];
            if (host.isPHost() && config.NumExperienceLevels() > 0) {
                int16_t level = sh.unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
                if (level >= 0 && level <= NUM_EXPERIENCE_LEVELS) {
                    ++levelCounts[level];
                    ++levels;
                }
            }
        }
    }

    /* Heading */
    d.add(RichText(UIColor::tc_White, withFreighters ? _("Starships") : _("Capital Ships"))
          .withAttribute(RichTextStyleAttribute::Big));
    d.addParagraph();

    /* Top part */
    const int em = font_ems[FONT_NORMAL];
    if ((opts & hide_top) == 0) {
        d.add(_("Total:"));
        d.addRight(22*em, RichText(UIColor::tc_Green, numToString(count)));
        d.addNewline();

        string_t querySuffix = " And Owner$=My.Race$";
        if (!withFreighters)
            querySuffix += " And Type.Short<>\"F\"";
        for (int i = 0; i < ITEMS; ++i) {
            if (counts[i] > 0) {
                d.add(RichText(_(names[i]))
                      .withAttribute(new RichTextLinkAttribute(format("q:UI.Search '%s%s', '2s'", exprs[i], querySuffix))));
                d.addRight(22*em, RichText(UIColor::tc_Green, numToString(counts[i])));
                d.addNewline();
            }
        }

        if (levels > 0) {
            d.addNewline();
            d.add(RichText(UIColor::tc_White, _("Ships by Experience Level")));
            d.addNewline();
            for (int i = 0; i <= NUM_EXPERIENCE_LEVELS; ++i) {
                if (levelCounts[i] > 0) {
                    d.add(RichText(getExperienceLevelName(i))
                          .withAttribute(new RichTextLinkAttribute(format("q:UI.Search 'Level=%d%s', '2s'", i, querySuffix))));
                    d.addRight(22*em, RichText(UIColor::tc_Green, numToString(levelCounts[i])));
                    d.addNewline();
                }
            }
        }
    }

    if (count == 0)
        return;

    d.addNewline();
    d.add(RichText(UIColor::tc_White, _("Ships by Hull Type")));
    d.addNewline();

    drawHullList(d, hullCounts, opts & sort_mask, "q:UI.Search 'Owner$=My.Race$ And Hull$=%d', '2s'");
}

/***************************** Starchart Page ****************************/

static void
computeSize(std::vector<int>& xs, int wrap, int size[2])
{
    if (xs.empty()) {
        /* no planets */
        size[0] = size[1] = -1;
    } else {
        /* sort for initial (and, for non-wrap, only) estimate */
        std::sort(xs.begin(), xs.end());
        size[0] = xs.front();
        size[1] = xs.back();

        if (chart_conf.getMode() == GChartConfiguration::Wrapped) {
            /* With wrap, we probably have a cluster yielding maximum size
               in the above computation:
                     . . .. . ..     . .
                     |                 |
                     |                 |
                     min               max
               We're now looking for a new set
                     . . .. . ..     . .
                               |     |
                               |     |
                               max   min
               giving a better estimate about empire size. */
            int est = size[1] - size[0];
            for (std::vector<int>::size_type i = 1; i < xs.size(); ++i) {
                int t = xs[i-1] - xs[i] + wrap;
                if (t < est) {
                    size[0] = xs[i];
                    size[1] = xs[i-1];
                    est = t;
                }
            }
        }
    }
}

static void
showRange(RichDocument& d, string_t ttl, int size[], int wrap)
{
    const int em = font_ems[FONT_NORMAL];
    if (size[0] == size[1]) {
        d.add(string_t(format("%s Location:", ttl)));
        d.addAt(20*em, RichText(UIColor::tc_Green, format(_("at %d"), size[0])));
    } else {
        d.add(string_t(format("%s Range:", ttl)));
        if (size[0] < size[1]) {
            d.addAt(20*em, RichText(UIColor::tc_Green, format(_("%d ly from %d to %d"), size[1]-size[0]+1, size[0], size[1])));
        } else {
            d.addAt(20*em, RichText(UIColor::tc_Green, format(_("%d ly from %d to %d"), wrap+size[1]-size[0]+1, size[1], size[0])));
        }
    }
    d.addNewline();
}

static void
showIfNonzero(RichDocument& d, string_t ttl, int n)
{
    const int em = font_ems[FONT_NORMAL];
    if (n > 0) {
        d.add(ttl);
        d.addAt(20*em, RichText(UIColor::tc_Green, numToString(n)));
        d.addNewline();
    }
}

static int
sumForeign(const int array[])
{
    int sum = 0;
    for (int i = 1; i <= NUM_OWNERS; ++i)
        if (i != getPlayerId())
            sum += array[i];
    return sum;
}

static void
drawStarchartPage(RichDocument& d, GUniverse& univ)
{
    /* Acquire data. Totally different than in 1.x. */
    enum {
        CurrentPlanet,
        OldPlanet,
        CurrentShip,
        OldShip,
        Minefield,
        NumCounters
    };
    int byOwner[NumCounters][NUM_OWNERS+1];
    int umfPlanet = 0;
    int totalPlanets = 0;
    int totalTargets = 0;
    std::vector<int> planetXs;
    std::vector<int> planetYs;

    std::fill_n(byOwner[0], NumCounters*(NUM_OWNERS+1), 0);

    /* Planets */
    GAnyPlanetType& ptype = univ.ty_any_planets;
    for (int pid = ptype.findNextIndex(0); pid != 0; pid = ptype.findNextIndex(pid)) {
        const GPlanet& pl = ptype.getObjectByIndex(pid);
        /* Accout total */
        ++totalPlanets;

        /* Account played planets */
        if (pl.isPlayable(GObject::Playable)) {
            planetXs.push_back(pl.getPos().x);
            planetYs.push_back(pl.getPos().y);
            if (pl.getFCode().substr(0, 2) == "mf") {
                umfPlanet = pid;
            }
        }

        /* Account owners */
        if (pl.isOwnerKnown()) {
            int o = pl.getOwner();
            if (o >= 0 && o <= NUM_OWNERS) {
                if (pl.getHistoryTimestamp(ts_Colonists) == univ.getTurnNumber()) {
                    ++byOwner[CurrentPlanet][o];
                } else {
                    ++byOwner[OldPlanet][o];
                }
            }
        }
    }

    /* Ships */
    GAnyShipType& stype = univ.ty_any_ships;
    for (int sid = stype.findNextIndex(0); sid != 0; sid = stype.findNextIndex(sid)) {
        const GShip& sh = stype.getObjectByIndex(sid);
        int o = sh.getOwner();
        if (o >= 0 && o <= NUM_OWNERS) {
            if (sh.getShipKind() == GShip::HistoryShip || sh.getShipKind() == GShip::GuessedShip) {
                ++byOwner[OldShip][o];
            } else {
                ++byOwner[CurrentShip][o];
                if (sh.getShipKind() == GShip::CurrentTarget) {
                    ++totalTargets;
                }
            }
        }
    }

    /* Minefields */
    int ownMines = 0, teamMines = 0, enemyMines = 0;
    const GMinefieldType& mfs = univ.ty_minefields;
    for (GMinefieldType::const_iterator i = mfs.begin(); i != mfs.end(); ++i) {
        int o = i->getOwner();
        if (o >= 0 && o <= NUM_OWNERS) {
            ++byOwner[Minefield][o];
            switch (getPlayerRelation(o)) {
             case is_Me:
                ++ownMines;
                break;
             case is_Ally:
                ++teamMines;
                break;
             default:
                ++enemyMines;
                break;
            }
        }
    }

    /* Compute size */
    int sizeX[2], sizeY[2];
    computeSize(planetXs, chart_conf.getSize().x, sizeX);
    computeSize(planetYs, chart_conf.getSize().y, sizeY);

    /* Now, draw it */
    const int em = font_ems[FONT_NORMAL];
    d.add(RichText(UIColor::tc_White, _("Starchart")).withAttribute(RichTextStyleAttribute::Big));
    d.addParagraph();

    d.add(RichText(UIColor::tc_White, _("Your Empire")));
    d.addNewline();
    d.add(_("Planets:"));
    d.addAt(20*em, RichText(UIColor::tc_Green, numToString(byOwner[CurrentPlanet][::getPlayerId()])));
    d.addNewline();
    if (sizeX[0] >= 0) {
        d.setLeftMargin(em);
        showRange(d, _("East-West"), sizeX, chart_conf.getSize().x);
        showRange(d, _("North-South"), sizeY, chart_conf.getSize().y);
        d.setLeftMargin(0);
    }
    d.add(_("Starships:"));
    d.addAt(20*em, RichText(UIColor::tc_Green, numToString(byOwner[CurrentShip][::getPlayerId()])));
    d.addNewline();
    showIfNonzero(d, _("Unowned Planets:"), byOwner[CurrentPlanet][0] + byOwner[OldPlanet][0]);
    d.add(_("Total Planets:"));
    d.addAt(20*em, RichText(UIColor::tc_Green, numToString(totalPlanets)));
    d.addNewline();
    showIfNonzero(d, _("Ion Storms:"), univ.ty_ion_storms.countObjects());
    showIfNonzero(d, _("Own Minefields:"), ownMines);
    showIfNonzero(d, _("Team Minefields:"), teamMines);
    showIfNonzero(d, _("Enemy Minefields:"), enemyMines);
    d.addNewline();

    d.add(RichText(UIColor::tc_White, _("Foreign Units")));
    d.addRight(20*em, _("(ships)"));
    d.addRight(28*em, _("(planets)"));
    d.addRight(36*em, _("(minefields)"));
    d.addNewline();

    for (int pl = 1; pl <= NUM_OWNERS; ++pl) {
        if (pl != ::getPlayerId()
            && (byOwner[CurrentPlanet][pl] != 0
                || byOwner[OldPlanet][pl] != 0
                || byOwner[CurrentShip][pl] != 0
                || byOwner[OldShip][pl] != 0
                || byOwner[Minefield][pl] != 0))
        {
            d.add(RichText(player_racenames.getShortName(pl))
                  .withAttribute(new RichTextLinkAttribute(format("q:UI.Search 'Owner$=%d', '2pso'", pl))));
            d.addRight(18*em, RichText(UIColor::tc_Green, numToString(byOwner[CurrentShip][pl])));
            if (int n = byOwner[OldShip][pl])
                d.add(RichText(UIColor::tc_Green, format(" (+%d)", n)));
            d.addRight(26*em, RichText(UIColor::tc_Green, numToString(byOwner[CurrentPlanet][pl])));
            if (int n = byOwner[OldPlanet][pl])
                d.add(RichText(UIColor::tc_Green, format(" (+%d)", n)));
            d.addRight(34*em, RichText(UIColor::tc_Green, numToString(byOwner[Minefield][pl])));
            d.addNewline();
        }
    }

    d.add(_("Total:"));
    d.addRight(18*em, RichText(UIColor::tc_Green, numToString(sumForeign(byOwner[CurrentShip]))));
    if (int n = sumForeign(byOwner[OldShip]))
        d.add(RichText(UIColor::tc_Green, format(" (+%d)", n)));
    d.addRight(26*em, RichText(UIColor::tc_Green, numToString(sumForeign(byOwner[CurrentPlanet]))));
    if (int n = sumForeign(byOwner[OldPlanet]))
        d.add(RichText(UIColor::tc_Green, format(" (+%d)", n)));
    d.addRight(34*em, RichText(UIColor::tc_Green, numToString(sumForeign(byOwner[Minefield]))));
    d.addNewline();
    d.addAt(em, _("Visual Contacts:"));
    d.addRight(18*em, RichText(UIColor::tc_Green, numToString(totalTargets)));
    d.addNewline();

    d.addNewline();
    d.add(_("Universal Minefield FCode: "));
    if (umfPlanet) {
        d.add(RichText(UIColor::tc_Green, ptype.getObjectByIndex(umfPlanet).getFCode())
              .withAttribute(new RichTextLinkAttribute(format("q:UI.GotoScreen 2,%d", umfPlanet))));
    } else {
        d.add(RichText(UIColor::tc_Green, _("none")));
    }
}

/****************************** Weapon page ******************************/

static void
drawWeaponPage(RichDocument& d, GUniverse& univ, int opts)
{
    CTASSERT(NUM_BEAMS == NUM_TORPS);
    static const char link_prefix[] = "Owner$=My.Race$ And ";
    enum {
        TorpedoShips,
        Torpedoes,
        BeamShips,
        Beams,
        NumCounters
    };
    int32_t counters[NumCounters][NUM_BEAMS+1];
    int32_t unarmed = 0;
    int32_t carriers = 0;
    int32_t fighters = 0;
    std::fill_n(counters[0], NumCounters * (NUM_BEAMS+1), 0);

    GPlayedShipType& type = univ.ty_played_ships;
    for (int sid = type.findNextIndex(0); sid != 0; sid = type.findNextIndex(sid)) {
        const GShip& sh = type.getObjectByIndex(sid);
        if (sh.getNumBays() == 0 && sh.getNumBeams() == 0 && sh.getNumTorpLaunchers() == 0) {
            // Freighter
            ++unarmed;
            ++counters[TorpedoShips][0];
            ++counters[BeamShips][0];
        } else {
            // Primary weapon
            if (sh.getNumBeams() > 0 && sh.getBeamType() > 0 && sh.getBeamType() <= NUM_BEAMS) {
                ++counters[BeamShips][sh.getBeamType()];
                counters[Beams][sh.getBeamType()] += sh.getNumBeams();
            } else {
                ++counters[BeamShips][0];
            }
            // Secondary weapon
            if (sh.getNumTorpLaunchers() > 0 && sh.getTorpType() > 0 && sh.getTorpType() <= NUM_TORPS) {
                ++counters[TorpedoShips][sh.getTorpType()];
                counters[Torpedoes][sh.getTorpType()] += sh.getAmmo();
            } else if (sh.getNumBays() > 0) {
                ++carriers;
                fighters += sh.getAmmo();
            } else {
                ++counters[TorpedoShips][0];
            }
        }
    }

    /* Display it */
    const int em = font_ems[FONT_NORMAL];
    d.add(RichText(UIColor::tc_White, _("Weapons")).withAttribute(RichTextStyleAttribute::Big));
    d.addParagraph();

    if (opts == 0 || opts == show_only_Beams) {
        d.add(RichText(UIColor::tc_White, _("Beams")));
        d.addRight(20*em, _("Ships"));
        d.addRight(30*em, _("Weapons"));
        d.addNewline();
        for (int i = 0; i <= NUM_BEAMS; ++i) {
            if (opts != 0 || counters[BeamShips][i] != 0) {
                RichText str(i == 0 ? _("No beams") : getBeam(i).getName());
                if (counters[BeamShips][i])
                    str.withAttribute(new RichTextLinkAttribute(format("q:UI.Search '%sBeam$=%d','2s'", link_prefix, i)));
                d.add(str);
                d.addRight(20*em, RichText(UIColor::tc_Green, numToString(counters[BeamShips][i])));
                if (i != 0) {
                    d.addRight(30*em, RichText(UIColor::tc_Green, numToString(counters[Beams][i])));
                }
                d.addNewline();
            }
        }
        d.addNewline();
    }

    if (opts == 0 || opts == show_only_Torps) {
        d.add(RichText(UIColor::tc_White, _("Torpedoes")));
        d.addRight(20*em, _("Ships"));
        d.addRight(30*em, _("Torpedoes"));
        d.addNewline();
        if (opts != 0 || counters[TorpedoShips][0] != 0) {
            RichText str(_("No torps/fighters"));
            if (counters[TorpedoShips][0])
                str.withAttribute(new RichTextLinkAttribute(format("q:UI.Search '%sIsEmpty(Aux)','2s'", link_prefix)));
            d.add(str);
            d.addRight(20*em, RichText(UIColor::tc_Green, numToString(counters[TorpedoShips][0])));
            d.addNewline();
        }
        for (int i = 1; i <= NUM_BEAMS; ++i) {
            if (opts != 0 || counters[TorpedoShips][i] != 0) {
                RichText str(getTorp(i).getName());
                if (counters[TorpedoShips][i])
                    str.withAttribute(new RichTextLinkAttribute(format("q:UI.Search '%sTorp$=%d','2s'", link_prefix, i)));
                d.add(str);
                d.addRight(20*em, RichText(UIColor::tc_Green, numToString(counters[TorpedoShips][i])));
                d.addRight(30*em, RichText(UIColor::tc_Green, numToString(counters[Torpedoes][i])));
                d.addNewline();
            }
        }
        d.addNewline();
    }

    if (opts == 0 || opts == show_only_Rest) {
        d.add(RichText(UIColor::tc_White, _("Others")));
        d.addNewline();

        RichText str(_("Carriers"));
        if (carriers)
            str.withAttribute(new RichTextLinkAttribute(format("q:UI.Search '%sFighter.Bays','2s'", link_prefix)));
        d.add(str);
        d.addRight(20*em, RichText(UIColor::tc_Green, numToString(carriers)));
        d.addNewline();
        if (carriers) {
            d.addAt(em, _("Fighters"));
            d.addRight(20*em, RichText(UIColor::tc_Green, numToString(fighters)));
            d.addNewline();
        }

        str = _("Unarmed ships");
        if (unarmed)
            str.withAttribute(new RichTextLinkAttribute(format("q:UI.Search '%sType.Short=\"F\"','2s'", link_prefix)));
        d.add(str);
        d.addRight(20*em, RichText(UIColor::tc_Green, numToString(unarmed)));
        d.addNewline();
    }
}

/*********************** WImperialStatisticsWindow ***********************/

namespace {
    class WImperialStatisticsWindow : public UIWindow, public IntUserInterfaceDialogBinding {
        ObjectHolder h;
        UIRichDocument* doc;
        int page;

        void init();
        void setPage(int n);
     public:
        WImperialStatisticsWindow();

        bool handleEvent(const UIEvent& event, bool second_pass);
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        void drawPage(int n);
        void onLinkClick(string_t link);
        void doOptions(UIBaseWidget& sender);
    };
}

WImperialStatisticsWindow::WImperialStatisticsWindow()
    : UIWindow(_("Imperial Statistics"), BLUE_WINDOW, COLOR_DARK),
      doc(0),
      page(lastActivePage)
{
    init();
}

void
WImperialStatisticsWindow::init()
{
    // HBox
    //   UIRichDocument
    //   VBox
    //     UIButton "Totals"
    //     ...
    //     UISpacer
    //     HBox
    //       "Help"
    //       UISpacer
    //       "Close"

    setLayout(UIHBoxLayout::instance5);

    doc = &h.add(new UIRichDocument(GfxPoint(font_sizes[FONT_NORMAL].scaledBy(10, 10)), UIRichDocument::fl_SingleHyper));
    add(*doc);

    UIGroup& g1 = h.add(new UIGroup(UIVBoxLayout::instance5));
    g1.add(h.add(new UIButton(1000+TotalsPage,    UIKeyString(_("T - Totals")))).addCommand());
    g1.add(h.add(new UIButton(1000+MineralsPage,  UIKeyString(_("M - Minerals")))).addCommand());
    g1.add(h.add(new UIButton(1000+PlanetsPage,   UIKeyString(_("P - Planets")))).addCommand());
    g1.add(h.add(new UIButton(1000+ColonyPage,    UIKeyString(_("O - Colony")))).addCommand());
    g1.add(h.add(new UIButton(1000+StarbasePage,  UIKeyString(_("B - Starbases")))).addCommand());
    g1.add(h.add(new UIButton(1000+StarshipPage,  UIKeyString(_("S - Starships")))).addCommand());
    g1.add(h.add(new UIButton(1000+CapitalPage,   UIKeyString(_("C - Capital")))).addCommand());
    g1.add(h.add(new UIButton(1000+StarchartPage, UIKeyString(_("A - Starchart")))).addCommand());
    g1.add(h.add(new UIButton(1000+WeaponsPage,   UIKeyString(_("W - Weapons")))).addCommand());
    g1.add(h.add(new UIFixedSpacer(GfxPoint(10,10))));
    g1.add(h.add(new UIButton(2000, UIKeyString(_("# - Options")))).addCommand());
    g1.add(h.add(new UISpacer()));
    
    UIGroup& g11 = h.add(new UIGroup(UIHBoxLayout::instance5));
    g11.add(h.add(new UIButton(_("Help"), 'h')));
    g11.add(h.add(new UISpacer()));
    g11.add(h.add(new UIButton(1, _("Close"), SDLK_ESCAPE)).addStop());
    g1.add(g11);
    add(g1);
    add(h.add(new WHelpWidget("pcc2:imperial")));

    for (int i = 0; i < LastPage; ++i)
        setStateOf(1000+i, UIButton::bf_LeftJust, true);
    setStateOf(1000+page, UIButton::bf_Highlight, true);
    setStateOf(2000, UIButton::bf_LeftJust, true);

    doc->sig_link_click.add(this, &WImperialStatisticsWindow::onLinkClick);

    setExtent(ui_root->getExtent());

    drawPage(page);
}

void
WImperialStatisticsWindow::setPage(int n)
{
    if (n != page) {
        setStateOf(1000+page, UIButton::bf_Highlight, false);
        page = n;
        setStateOf(1000+page, UIButton::bf_Highlight, true);
        doc->getDocument().clear();
        drawPage(n);
        doc->handleDocumentUpdate();
        lastActivePage = page;
    }
}

bool
WImperialStatisticsWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_DOWN:
         case SDLK_PAGEDOWN:
            if (page < LastPage-1) {
                setPage(page+1);
            } else {
                setPage(0);
            }
            return true;
         case SDLK_UP:
         case SDLK_PAGEUP:
            if (page > 0) {
                setPage(page-1);
            } else {
                setPage(LastPage-1);
            }
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

bool
WImperialStatisticsWindow::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code >= 1000 && cmd.code < 1000+LastPage) {
        setPage(cmd.code - 1000);
        return true;
    } else if (cmd.code == 2000) {
        doOptions(sender);
        return true;
    } else {
        return UIWindow::handleCommand(cmd, sender);
    }
}

void
WImperialStatisticsWindow::drawPage(int n)
{
    RichDocument& d = doc->getDocument();
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    switch (n) {
     case TotalsPage:
        drawTotalsPage(d, univ);
        setStateOf(2000, st_Disabled, true);
        break;
     case MineralsPage:
        drawMineralsPage(d, univ, pageOptions[n]);
        setStateOf(2000, st_Disabled, false);
        break;
     case PlanetsPage:
        drawPlanetsPage(d, univ, pageOptions[n]);
        setStateOf(2000, st_Disabled, false);
        break;
     case ColonyPage:
        drawColonyPage(d, univ, pageOptions[n]);
        setStateOf(2000, st_Disabled, false);
        break;
     case StarbasePage:
        drawStarbasePage(d, univ, pageOptions[n]);
        setStateOf(2000, st_Disabled, false);
        break;
     case StarshipPage:
     case CapitalPage:
        drawStarshipPage(d, univ, pageOptions[n], n == StarshipPage);
        setStateOf(2000, st_Disabled, false);
        break;
     case StarchartPage:
        drawStarchartPage(d, univ);
        setStateOf(2000, st_Disabled, true);
        break;
     case WeaponsPage:
        drawWeaponPage(d, univ, pageOptions[n]);
        setStateOf(2000, st_Disabled, false);
        break;
     default:
        break;
    }
    d.finish();
}

void
WImperialStatisticsWindow::onLinkClick(string_t link)
{
    if (link.size() > 2 && link.compare(0, 2, "q:", 2) == 0) {
        executeStatement(link.substr(2), 0);
    }
}

static void
addSortOrders(StringList& list, int hi)
{
    list.add(hi + sort_by_id,      _("Sort ships by hull Id"));
    list.add(hi + sort_by_name,    _("Sort ships by name"));
    list.add(hi + sort_by_amount,  _("Sort ships by count"));
    list.add(hi + sort_by_mass,    _("Sort ships by hull mass"));
    list.add(hi + sort_by_tech,    _("Sort ships by tech level"));
    list.add(hi + sort_by_cargo,   _("Sort ships by cargo room"));
    list.add(hi + sort_by_engines, _("Sort ships by number of engines"));
    list.add(hi + sort_by_beams,   _("Sort ships by maximum beams"));
    list.add(hi + sort_by_tubes,   _("Sort ships by maximum torpedo launchers"));
    list.add(hi + sort_by_bays,    _("Sort ships by number of fighter bays"));
}

void
WImperialStatisticsWindow::doOptions(UIBaseWidget& sender)
{
    StringList list;
    int hi = pageOptions[page] & 0xF0;
    int lo = pageOptions[page] & 0x0F;
    switch (page) {
     case PlanetsPage:
        list.add(sort_P_by_Race, _("Sort by native race"));
        list.add(sort_P_by_NumPlanets, _("Sort by number of planets"));
        list.add(sort_P_by_TotalPop, _("Sort by total population"));
        break;
     case StarshipPage:
     case CapitalPage:
        addSortOrders(list, hi);
        if (hi) {
            list.add(lo, _("Show all info"));
        } else {
            list.add(lo + hide_top, _("Show only hull list"));
        }
        break;
     case StarbasePage:
        addSortOrders(list, hi);
        if (hi) {
            list.add(lo, _("Show all info"));
        } else {
            list.add(lo + hide_top, _("Show only ship list"));
        }
        break;
     case MineralsPage:
        list.add(hi + sort_by_total,  _("Sort by total amount"));
        list.add(hi + sort_by_mined,  _("Sort by mined amount"));
        list.add(lo + show_only_N,    _("Show only Neutronium"));
        list.add(lo + show_only_T,    _("Show only Tritanium"));
        list.add(lo + show_only_D,    _("Show only Duranium"));
        list.add(lo + show_only_M,    _("Show only Molybdenum"));
        list.add(lo,                  _("Show all 4 minerals"));
        break;
     case ColonyPage:
        list.add(lo, _("Show all info"));
        list.add(lo + show_only_Colonists, _("Show only Colonists"));
        list.add(lo + show_only_Supplies,  _("Show only Supplies"));
        list.add(lo + show_only_Money,     _("Show only Money"));
        break;
     case WeaponsPage:
        list.add(show_only_Beams, _("Show only beams"));
        list.add(show_only_Torps, _("Show only torpedoes"));
        list.add(show_only_Rest,  _("Show only rest"));
        list.add(show_everything, _("Show all info"));
        break;
    }

    if (list.getCount() > 0) {
        UIStandardListbox lb(0, 1, getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
        lb.scrollToKey(pageOptions[page]);
        if (lb.doStandardMenu(sender)) {
            pageOptions[page] = lb.getKey();
            doc->getDocument().clear();
            drawPage(page);
            doc->drawWidget();
        }
    }
}

void
doImperialStatistics()
{
    WImperialStatisticsWindow win;
    win.runModal(win);
}

/***************************** Initialisation ****************************/

static void
initImperialStatistics()
{
    std::fill_n(&pageOptions[0], std::size_t(LastPage), 0);
    lastActivePage = TotalsPage;
}

const ModuleInitializer mod_imperial[] = {
    { OnEnterGame | OnLeftGame, initImperialStatistics },
    { 0, 0 },
};
