/**
  *  \file flak/pdk-play.cc
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <phostpdk.h>
#include "io/streams.h"
#include "util/except.h"
#include "util/console.h"
#include "util/modinit.h"
#include "flak/flak.h"

static const char* filename;
static const char* progname;

static void
help()
{
    ErrorExit("usage: %s FILE [GAMEDIR [ROOTDIR]]", progname);
}

static void
play(FlakBattle& b, const TFlakBattle& h)
{
    FlakNullVisualizer vis(b);
    Info("  Time according to header: %7d", h.total_time);
    b.init();
    while (!b.playOneTick())
        ;
    Info("  Real time taken:          %7d", b.getTime());
    for (int i = 0; i < b.getNumShips(); ++i) {
        FlakShip& sh = b.getShipByNumber(i);
        Info("    Unit %3d (%-6s #%-3d): damage %3d, crew %4d, shield %3d, torps %3d, fighters %3d",
             i,
             (sh.data.flags & flak_IsPlanet) ? "planet" : "ship",
             sh.data.id,
             int(sh.status.damage),
             int(sh.status.crew),
             int(sh.status.shield),
             sh.status.torp_count,
             sh.status.fighter_count);
    }
}

int
main(int /*argc*/, char** argv)
{
    bool had_gamedir = false, had_rootdir = false;
    progname = argv[0];
    while (const char* p = std::strpbrk(progname, ":/\\"))
        progname = p+1;
    while (*++argv) {
        if (filename == 0)
            filename = *argv;
        else if (!had_gamedir)
            gGameDirectory = *argv, had_gamedir = true;
        else if (!had_rootdir)
            gRootDirectory = *argv, had_rootdir = true;
        else
            help();
    }
    if (!filename)
        help();

    // performInitialisation(OnStartup, mod_base_console);
    InitPHOSTLib();
    if (!ReadGlobalData())
        ErrorExit("Unable to read global data");
//     if (!Read_HConfig_File())
//         ErrorExit("Unable to read host config");
    if (!Read_HConfig_File())
        ErrorExit("Unable to read host data");

    try {
        /* Now read the input file */
        Ptr<Stream> io = Stream::openFileNT(filename, Stream::C_READ);
        if (!io)
            ErrorExit("Unable to open file %s", filename);

        /* We can no longer use GFlakVcrDatabase, so do it manually */
        TFlakHeader header;
        getStructureT(*io, header);
        if (std::memcmp(header.magic, FLAK_MAGIC, sizeof(FLAK_MAGIC)) != 0)
            throw FileFormatException(*io, "File is missing required signature");
        if (header.filefmt_version != 0)
            throw FileFormatException(*io, "Unsupported file format version");

        for (int i = 0; i < header.num_battles; ++i) {
            /* read buffer */
            Buffer<char> data;
            data.ensureSize(4);
            io->readT(data.ptr(), 4);
            uint32_t size = getUint32(data.ptr());
            data.ensureSize(size);
            io->readT(data.ptr()+4, size-4);

            /* make battle */
            FlakBattle b;
            TFlakBattle h;
            b.loadFromStream(filename, data, h);

            /* play it */
            Info("Battle %d...", i+1);
            play(b, h);
        }
    }
    catch (std::exception& e) {
        ErrorExit("Exception: %s", e.what());
    }
    catch (...) {
        ErrorExit("Unknown exception");
    }
    return 0;
}
