/**
  *  \file game/drawing.cc
  *  \brief User Drawings
  *
  *  Classes to represent and manage user-created starchart drawings.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <cmath>
#include "game/drawing.h"
#include "cpluslib/assert.h"
#include "util/console.h"
#include "cpluslib/buffer.h"
#include "util/nls.h"
#include "game/struct.h"
#include "util/except.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/recode.h"

/**************************** Color Conversion ***************************/

/* PCC 1.x has its color map set up as
     $00: <16 colors 'A'>
     $10: <16 colors 'A'>
     $20: <16 colors 'A'>
     $30: <80 colors 'B'>
     $80: <16 colors 'C'> <- these are available to users as colors 11-30
     $90: <48 colors 'D'>
     $C0: 64 available slots
   We have condensed this as
     $00: <16 colors 'A'>
     $10: <80 colors 'B'>
     $60: <16 colors 'C'>
     $70: <48 colors 'D'>
     $A0: 96 available slots
   (see palette.h). We provide a 1:1 mapping between the two sequences, such
   that every value stored by us comes out at the intended value, and we read
   PCC 1.x user colors correctly. */

/** Map external color representation to internal one.
    \inv convertExternalColorToInternal(convertInternalColorToExternal(x)) == x */
static uint8
convertExternalColorToInternal(uint8 value)
{
    if (value < 0x10)
        /* keep first 'A' range */
        return value;
    else if (value < 0x30)
        /* map second and third 'A' range to available range; this does not
           appear in real files */
        return value - 0x10 + 0xA0;
    else if (value < 0xC0)
        /* map B,C,D ranges */
        return value - 0x20;
    else
        /* keep available range; this does not appear in real files */
        return value;
}

/** Map internal color representation to external one.
    \inv convertInternalColorToExternal(convertExternalColorToInternal(x)) == x */
static uint8
convertInternalColorToExternal(uint8 value)
{
    if (value < 0x10)
        return value;
    else if (value < 0xA0)
        return value + 0x20;
    else if (value < 0xC0)
        return value - 0xA0 + 0x10;
    else
        return value;
}

/* In addition to the internal<->external conversion, we expose a
   simple color number (0..NUM_USER_COLORS, where 0 isn't selectable
   through dialogs) to the user. The following two functions provide a
   mapping. */
static const uint8 user_colors[] = {
    0,
    1,   2,   3,   4,   5,   6,   7,   8,   9,  15,
    97,  99,  101, 103, 105, 107, 109, 111, 113, 115,
    98,  100, 102, 104, 106, 108, 110, 112, 114, 116
};

CTASSERT(countof(user_colors) == GDrawing::NUM_USER_COLORS+1);

/** Convert color into user color index.
    \param color Color number
    \return User color number (0..NUM_USER_COLORS); -1 if color does not match
    any user color */
int
GDrawing::getUserColorFromColor(uint8 color)
{
    for (int i = 0; i < int(countof(user_colors)); ++i)
        if (user_colors[i] == color)
            return i;
    return -1;
}

/** Convert user color index into color number.
    \param user_color User color number (0..NUM_USER_COLORS)
    \return color number */
uint8
GDrawing::getColorFromUserColor(int user_color)
{
    if (user_color >= 0 && user_color < int(countof(user_colors)))
        return user_colors[user_color];
    return uint8(-1);
}

/** Check whether value approximately lies in range.
    \param testee [in] Value to test
    \param a,b    [in] Range
    \return true iff \c testee can be considered close to or between \c a and \c b */
static bool
isInRangeApprox(int testee, int a, int b)
{
    if (a > b)
        std::swap(a, b);

    return testee >= a - 20
        && testee <= b + 20;
}

/******************************** GDrawing *******************************/

/** Constructor. Make a new drawing.
    \param pos Position or center point.
    \param type Type of drawing.

    All other fields are set to defaults and should be set using the
    setXXX functions. */
GDrawing::GDrawing(GPoint pos, GDrawingType type)
    : pos(pos), type(type), color(1),
      x2(0), y2(0), tag(0), expire(-1), comment()
{ }

/** Set position.
    \param pos Position or center point. */
void
GDrawing::setPos(GPoint pos)
{
    this->pos = pos;
}

/** Set other position. Valid only for RectangleDrawing and LineDrawing.
    \param pos Position of bottom/right corner. */
void
GDrawing::setPos2(GPoint pos)
{
    ASSERT(type == RectangleDrawing || type == LineDrawing);
    this->x2 = pos.x;
    this->y2 = pos.y;
}

/** Set circle radius. Valid only for CircleDrawing.
    \param r New radius */
void
GDrawing::setCircleRadius(int r)
{
    ASSERT(type == CircleDrawing);
    this->x2 = r;
}

/** Set marker kind (shape). Valid only for MarkerDrawing.
    \param k New kind (see getUserMarker()) */
void
GDrawing::setMarkerKind(int k)
{
    ASSERT(type == MarkerDrawing);
    this->x2 = k;
}

/** Set marker tag. This is a general-purpose numeric value, conventionally
    an atom.
    \param tag New tag */
void
GDrawing::setTag(atom_t tag)
{
    this->tag = tag;
}

/** Set color. By convention, markers allow colors 0-9 and 15, as well
    as 97-116. 0 means invisible.
    \param color Color index (COLOR_xxx, see gfx/palette.h) */
void
GDrawing::setColor(uint8 color)
{
    this->color = color;
}

/** Set comment. The comment is limited to 255 characters for storage.
    \param comment New comment. */
void
GDrawing::setComment(string_t comment)
{
    this->comment = comment;
}

/** Set time of expiry. Specifies a turn number. When a turn after that
    is seen, the marker is deleted (not loaded).
    \param expire Turn of expiry, -1 for never */
void
GDrawing::setExpire(int expire)
{
    this->expire = expire;
}

/** Compute distance of this drawing to a given point in the game
    universe. This is used to select a drawing for editing. This is an
    almost straightforward port of the same code in PCC 1.x. It does,
    however, <em>not</em> special-case locked-on markers or check
    visibility / editability; that must be done by the caller. */
double
GDrawing::getDistanceTo(GPoint pt) const
{
    switch (type) {
     case LineDrawing:
        if (isInRangeApprox(pt.x, pos.x, x2) && isInRangeApprox(pt.y, pos.y, y2)) {
            /* position is approximately within bounding rectangle of line */
            long d0 = distanceSquaredRaw(pos, GPoint(x2, y2));
            if (d0 == 0) {
                /* degenerate case: line has length 0 */
                return std::sqrt(double(distanceSquared(pos, pt)));
            } else {
                /* regular case: compute distance between point and line */
                int32 det = int32(pt.y - pos.y) * int32(x2 - pos.x)
                          - int32(pt.x - pos.x) * int32(y2 - pos.y);
                return std::abs(det / std::sqrt(double(d0)));
            }
        } else {
            /* position is outside bounding rectangle, don't bother looking further */
            /* FIXME: only the Line case still has this hardwired restriction */
            return 1e6;
        }

     case RectangleDrawing:
        /* for rectangles, simply the distance to either edge */
        /* FIXME: what about wrap? */
        if (isInRangeApprox(pt.x, pos.x, x2) && isInRangeApprox(pt.y, pos.y, y2)) {
            return std::min(std::min(std::abs(pt.x - pos.x),
                                     std::abs(pt.x - x2)),
                            std::min(std::abs(pt.y - pos.y),
                                     std::abs(pt.y - y2)));
        } else {
            return 1e6;
        }

     case CircleDrawing:
        return std::abs(std::sqrt(double(distanceSquared(pos, pt))) - x2);

     case MarkerDrawing:
        return std::sqrt(double(distanceSquared(pos, pt)));

     default:
        return 1e6;
    }
}

/** Compute distance of this drawing to a given point in the game
    universe. Same as getDistanceTo, but honors wrapping. */
double
GDrawing::getDistanceToWrap(GPoint pt) const
{
    double dist = getDistanceTo(pt);
    for (int i = 1, n = chart_conf.getNumRectangularImages(); i < n; ++i) {
        GPoint tmp;
        if (chart_conf.getPointAlias(pt, tmp, i, false)) {
            double ndist = getDistanceTo(tmp);
            if (ndist < dist)
                dist = ndist;
        }
    }
    return dist;
}


/**************************** GDrawingAtomMap ****************************/

/** Create blank map. */
GDrawingAtomMap::GDrawingAtomMap()
    : atoms(),
      counter(EXTERNAL_ATOM_MIN)
{ }
              
GDrawingAtomMap::~GDrawingAtomMap()
{ }

/** Clear this map. The object will behave as if it had just been constructed.
    \post isEmpty() */
void
GDrawingAtomMap::clear()
{
    map_t().swap(atoms);
    counter = EXTERNAL_ATOM_MIN;
}

/** Check for emptiness. */
bool
GDrawingAtomMap::isEmpty() const
{
    return counter == EXTERNAL_ATOM_MIN;
}

/** Add new internal atom. Registers that \c a is used as an atom by a
    relevant drawing. */
void
GDrawingAtomMap::add(atom_t a)
{
    if (atoms.find(a) == atoms.end()) {
        atoms[a] = counter++;
    }
}

/** Save object to stream. This is used to create the rPaintingTags
    (11) record in chartX.cc. */
void
GDrawingAtomMap::save(Stream& out)
{
    // build value array
    Buffer<char> value_buffer;
    TInt16 count;
    count.value = 0;
    for (map_t::const_iterator i = atoms.begin(); i != atoms.end(); ++i) {
        if (count.value >= 16000) {
            // too big. PCC 1.x supports up to 64000 bytes for the whole
            // Atom Translation entry, and 16000 is a very optimistic
            // estimate how many fit in there.
            console.write(LOG_WARN, _("Too many different marker tags. Some were ignored."));
            break;
        }
        value_buffer.append(i->second & 255);
        value_buffer.append(i->second >> 8);
        ++count.value;
    }
    ASSERT(uint16(count.value) * 2UL == value_buffer.getSize());

    // write it out
    storeStructureT(out, count);
    out.writeT(value_buffer.ptr(), value_buffer.getSize());

    int failed_strings = 0;
    for (uint16 i = 0; i < count.value; ++i) {
        const atom_t internal_atom = atoms[getInt16(value_buffer, 2L*i)];
        const string_t atom_name = convertUtf8ToGame(atomStr(internal_atom));
        if (atom_name.size() > 255)
            ++failed_strings;
        storePascalStringT(out, atom_name);
    }

    if (failed_strings != 0)
        console.write(LOG_WARN, format(_("%d marker tag%!1{s%} were too long to be stored in the chart file, and were truncated to 255 characters."),
                                       failed_strings));
}

/** Save object from stream. This is used to load the rPaintingTags
    (11) record from chartX.cc. */
void
GDrawingAtomMap::load(Stream& in)
{
    // FIXME: this does not maintain /counter/ and will thus yield an inconsistent state.
    // Maybe it makes sense to move this into a separate class where we can also make
    // a decent get() function?
    clear();

    // load 'count' field
    TInt16 count;
    getStructureT(in, count);
    if (count.value < 0)
        throw FileFormatException(in, _("Invalid value"));

    // load value array
    uint32 value_size = 2L * count.value;
    Buffer<char> value_buffer(value_size);
    value_buffer.ensureSize(value_size);
    in.readT(value_buffer.ptr(), value_size);

    // load strings and populate array
    for (int32 i = 0; i < count.value; ++i) {
        string_t external_string;
        getPascalStringT(in, external_string);
        int16 external_atom = getInt16(value_buffer, 2L*i);
        atom_t internal_atom = atom(convertGameToUtf8(external_string));
        atoms[internal_atom] = external_atom;
    }
}

/** Get iterator to first element. */
GDrawingAtomMap::iterator
GDrawingAtomMap::begin() const
{
    return atoms.begin();
}

/** Get iterator to one-past-last element. */
GDrawingAtomMap::iterator
GDrawingAtomMap::end() const
{
    return atoms.end();
}

/** Convert external atom to internal.
    \param value Value read from file
    \return equivalent atom to use internally */
atom_t
GDrawingAtomMap::get(int16 value) const
{
    // FIXME: we can surely do better
    for (map_t::const_iterator i = atoms.begin(); i != atoms.end(); ++i)
        if (i->second == value)
            return i->first;
    return atom_t(value);
}

/** Convert internal atom to external.
    \param value Value stored in GDrawing object
    \return equivalent value to store in file */
int16
GDrawingAtomMap::getExternalValue(atom_t atom) const
{
    map_t::const_iterator i = atoms.find(atom);
    if (i != atoms.end())
        return i->second;
    return int16(atom);
}


/*************************** GDrawingContainer ***************************/

/** Create blank container. */
GDrawingContainer::GDrawingContainer()
    : drawings()
{ }

GDrawingContainer::~GDrawingContainer()
{ }

/** Add new drawing.
    \param drawing Object, will be copied */
GDrawingContainer::iterator
GDrawingContainer::add(const GDrawing& drawing)
{
    iterator ret = drawings.insert(drawings.end(), new GDrawing(drawing));
    sig_changed.raise();
    return ret;
}

/** Compute atom map. Populates a GDrawingAtomMap object, used for
    chart file I/O and filtering.
    \param map   [out] Object to populate
    \param flags [in] Options, bitfield of IncludeNumeric, IncludeInvisible. */
void
GDrawingContainer::computeAtomMap(GDrawingAtomMap& map, int flags) const
{
    for (const_iterator i = begin(); i != end(); ++i) {
        const GDrawing& d = **i;
        if (!d.isVisible() && (flags & IncludeInvisible) == 0)
            /* invisibile marker */
            continue;

        atom_t a = d.getTag();
        if (!isAtom(a) && (flags & IncludeNumeric) == 0)
            /* numeric tag */
            continue;

        /* ok, we want this one */
        map.add(a);
    }
}

/** Save drawings in file
    \param out [in] Stream to write to
    \param map [in] Maps internal tags to external */
void
GDrawingContainer::save(Stream& out, const GDrawingAtomMap& map) const
{
    for (const_iterator i = begin(); i != end(); ++i) {
        /* Fetch drawing */
        const GDrawing& d = **i;

        /* Fetch values */
        string_t comment = convertUtf8ToGame(d.getComment());
        uint8_t kind = d.getType();
        if (comment.size())
            kind |= 0x80;

        /* Build structure */
        char marker_buffer[14];
        marker_buffer[0] = char(kind);
        marker_buffer[1] = char(convertInternalColorToExternal(d.getColor()));
        storeInt16(marker_buffer+2, d.pos.x);
        storeInt16(marker_buffer+4, d.pos.y);
        storeInt16(marker_buffer+6, d.x2);
        storeInt16(marker_buffer+8, d.y2);
        storeInt16(marker_buffer+10, map.getExternalValue(d.getTag()));
        storeInt16(marker_buffer+12, d.getExpire());
        out.writeT(marker_buffer, sizeof(marker_buffer));
        if (comment.size())
            storePascalStringT(out, comment);
    }
}

/** Load drawings from file.
    \param in     [in] Stream to read from (must yield EOF at end of data)
    \param map    [in] Maps external tags to internal
    \param expire [in] Current turn for expiration; -1 to not expire anything */
void
GDrawingContainer::load(Stream& in, const GDrawingAtomMap& map, const int expire)
{
    char marker_buffer[14];
    while (in.read(marker_buffer, sizeof(marker_buffer)) == sizeof(marker_buffer)) {
        /* Parse what we have so far: */
        const int kind             = uint8_t(marker_buffer[0]) & 127;
        const bool has_comment     = (uint8_t(marker_buffer[0]) & 0x80) != 0;
        const uint8_t color        = uint8_t(marker_buffer[1]);
        const int16_t x1           = getInt16(marker_buffer, 2);
        const int16_t y1           = getInt16(marker_buffer, 4);
        const int16_t x2           = getInt16(marker_buffer, 6);
        const int16_t y2           = getInt16(marker_buffer, 8);
        const int16_t external_tag = getInt16(marker_buffer, 10);
        const int16_t this_expire  = getInt16(marker_buffer, 12);

        /* If it has a comment, read that too */
        string_t comment;
        if (has_comment)
            getPascalStringT(in, comment);

        /* Is it expired? If yes, don't bother checking further */
        if (expire > 0 && this_expire >= 0 && this_expire < expire)
            continue;

        /* Check type */
        if (kind < 0 || kind > MarkerDrawing)
            continue;

        /* Might be valid */
        GDrawing t(GPoint(x1,y1), GDrawingType(kind));
        t.setColor(convertExternalColorToInternal(color));
        t.setTag(map.get(external_tag));
        t.setExpire(this_expire);
        t.setComment(convertGameToUtf8(comment));
        switch (kind) {
         case LineDrawing:
         case RectangleDrawing:
            t.setPos2(GPoint(x2,y2));
            break;
         case CircleDrawing:
            t.setCircleRadius(x2);
            break;
         case MarkerDrawing:
            t.setMarkerKind(x2);
            break;
        }
        add(t);
    }
}

/** Find nearest visible drawing.
    \param pt Location to start looking from
    \return iterator pointing to drawing, end() if no drawing can be considered near */
GDrawingContainer::iterator
GDrawingContainer::findNearestVisibleDrawing(GPoint pt)
{
    iterator found = end();
    double mindist = 21; /* minimum distance is less than 21 ly */
    for (iterator i = begin(), e = end(); i != e; ++i) {
        double dist = (*i)->getDistanceToWrap(pt);
        if (dist < mindist) {
            mindist = dist;
            found = i;
        }
    }
    return found;
}

/** Erase a drawing by iterator. */
void
GDrawingContainer::erase(iterator it)
{
    // Erase it
    sig_deleted.raise(it);
    drawings.erase(it);
}

/** Erase all expired drawings.
    \param turn Current turn number
    \todo We can actually do this during loading, when we know the current turn number
    there. Currently, we don't know it there. */
void
GDrawingContainer::eraseExpiredDrawings(int turn)
{
    for (iterator i = begin(), e = end(); i != e; ++i) {
        GDrawing& d = **i;
        if (d.getExpire() >= 0 && d.getExpire() < turn) {
            erase(i);
        }
    }
}
