/**
  *  \file game/save-trn.cc
  *  \brief Saving Turn Files
  *
  *  This module provides functions to save a game turn into a set of
  *  TRN files.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/format.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/reg.h"
#include "game/save.h"
#include "game/ship.h"
#include "game/trn.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/msg.h"
#include "util/recode.h"
#include "util/misc.h"
#include "game/timestamp.h"
#include "cpluslib/misc.h"
#include "game/racename.h"
#include "io/dirs.h"
#include "game/backup.h"
#include "util/progress.h"
#include "util/conf.h"

/** Send a single message into a turn file. */
static void
sendMessage(GTurnfile& trn, int from, int to, string_t text)
{
    encodeMessage(text);

    char buffer[4];
    storeUint16(buffer + 0, from);
    storeUint16(buffer + 2, to);

    trn.addCommand(tcm_SendMessage, text.length(), buffer, sizeof buffer);
    trn.addData(text.data(), text.length());
}

/** Send a message outbox into a turn file. */
static void
sendOutbox(GTurnfile& trn, GOutbox& outbox)
{
    for (int i = 0; i < outbox.getCount(); ++i) {
        GPlayerSet rec = outbox.getReceiverMask(i);
        for (int p = 1; p <= NUM_PLAYERS+1; ++p) {
            if (rec.contains(p))
                sendMessage(trn, outbox.getSender(), p,
                            outbox.getSendPrefix(i, p) + outbox.getRawText(i));
        }
    }
}

/** Send THost alliance commands. For THost, the mere presence of a
    friendly code change command triggers an alliance action. When we
    change the FCode back to what it should be, we can set alliances
    without sacrificing a ship for each action. */
static void
sendTHostAllies(GTurnfile& trn, const GCommand& cmd, const GShip& ship)
{
    char buffer[3];

    string_t s = convertUtf8ToGame(cmd.getArg());
    while (s.length() > 0) {
        storeBasicString(buffer, s);
        trn.addCommand(tcm_ShipChangeFc, ship.getId(), buffer, 3);
        if (s.length() <= 3)
            break;
        s.erase(0, 3);
    }
    storeBasicString(buffer, convertUtf8ToGame(ship.getFCode()));
    trn.addCommand(tcm_ShipChangeFc, ship.getId(), buffer, 3);
}

/** Save turn files (TRNs). This will generate all turns at once.
    \param trn          data base
    \param dir          target directory
    \param player_mask  whose turns to generate (1 << pid) */
void
saveTurns(GGameTurn& trn, Directory& dir, GPlayerSet player_mask)
{
    Ptr<GTurnfile> trns[NUM_PLAYERS];

    if (player_mask.empty())
        return;

    /* FIXME: this kind of timestamp handling is a hack. We need a
       unique timestamp to obtain the templock block, but we don't
       know whether all our timestamps actually are the same. This
       can't cause bugs, though, only templock blocks carried over
       too long. */
    GTimestamp timestamp;

    console.write(LOG_INFO, _("Generating turn commands..."));
    for (int pid = 1; pid <= NUM_PLAYERS; ++pid) {
        if (!player_mask.contains(pid))
            continue;

        /* red tape */
        GGen& gen = trn.getGen(pid);
        timestamp = gen.getTimestamp();
        trns[pid-1] = new GTurnfile(pid, timestamp);
        trns[pid-1]->setFeatures(GTurnfile::trnf_Winplan);
        trns[pid-1]->setRegInfo(::getRegInfo(), trn.getTurnNumber());

        /* make commands */
        int tha_ship_id = 0;
        const GUniverse& u = trn.getCurrentUniverse();
        const GUniverse& prev = trn.getPreviousUniverse();
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            if (u.getShip(i).isVisible() && u.getShip(i).getShipSource().contains(pid)) {
                if (tha_ship_id == 0)
                    tha_ship_id = i;

                // Get ship data
                TShip datShip, disShip;
                u.getShip(i).getShipData(datShip);
                prev.getShip(i).getShipData(disShip);

                // Convert into blobs and figure out differences
                char datBlob[TShip::size], disBlob[TShip::size];
                storeStructure(datBlob, datShip);
                storeStructure(disBlob, disShip);
                trns[pid-1]->makeShipCommands(i, disBlob, datBlob);
            }
        }
        for (int i = 1; i <= NUM_PLANETS; ++i) {
            if (u.getPlanet(i).isVisible() && u.getPlanet(i).getPlanetSource().contains(pid)) {
                // Get planet data
                TPlanet datPlanet, disPlanet;
                u.getPlanet(i).getPlanetData(datPlanet);
                prev.getPlanet(i).getPlanetData(disPlanet);

                // Convert into blobs and figure out differences
                char datBlob[TPlanet::size], disBlob[TPlanet::size];
                storeStructure(datBlob, datPlanet);
                storeStructure(disBlob, disPlanet);
                trns[pid-1]->makePlanetCommands(i, disBlob, datBlob);
            }
        }
        for (int i = 1; i <= NUM_PLANETS; ++i) {
            if (u.getPlanet(i).isVisible() && u.getPlanet(i).hasBase() && u.getPlanet(i).getBaseSource().contains(pid)) {
                // Get starbase data
                TStarbase datBase, disBase;
                u.getPlanet(i).getBaseData(datBase);
                prev.getPlanet(i).getBaseData(disBase);

                // Convert into blobs and figure out differences
                char datBlob[TStarbase::size], disBlob[TStarbase::size];
                storeStructure(datBlob, datBase);
                storeStructure(disBlob, disBase);
                trns[pid-1]->makeBaseCommands(i, disBlob, datBlob);
            }
        }

        /* messages */
        sendOutbox(*trns[pid-1], trn.getOutbox(pid));

        /* command messages */
        const GCommandContainer& cms = trn.getCommands(pid);
        string_t accum;
        for (GCommandContainer::const_iterator i = cms.begin(); i != cms.end(); ++i) {
            GCommand& cmd = **i;
            if (cmd.getCommand() == GCommand::phc_TAlliance) {
                if (!tha_ship_id)
                    console.write(LOG_WARN, format(_("Player %d has no ship; alliance changes not transmitted"), pid));
                else
                    sendTHostAllies(*trns[pid-1], cmd, trn.getCurrentUniverse().getShip(tha_ship_id));
            } else {
                string_t text = cmd.getCommandText();
                if (text.length() && text[0] != '$') {
                    if (accum.length() + text.length() > 500) {
                        sendMessage(*trns[pid-1], pid, pid, accum);
                        accum.clear();
                    }
                    if (accum.length())
                        accum += '\n';
                    accum += text;
                }
            }
        }
        if (accum.length())
            sendMessage(*trns[pid-1], pid, pid, accum);

        /* password */
        char new_password[10];
        if (gen.getNewPasswordData(new_password)) {
            trns[pid-1]->addCommand(tcm_ChangePassword, 0, new_password, sizeof(new_password));
        }

        /* generate turn */
        trns[pid-1]->update();
    }

    /* update trailers */
    uint32 cksums[NUM_PLAYERS];
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        Ptr<Stream> file = dir.openFileNT(format("player%d.trn", i), Stream::C_READ);
        if (file) {
            try {
                GTurnfile trn(*file, false);
                if (trn.getTimestamp() == timestamp) {
                    copy_n(trn.getDosTrailer().players, countof(cksums), cksums);
                    goto have_block;
                }
            }
            catch(...) { }
        }
    }

    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (trns[i-1])
            cksums[i-1] = trns[i-1]->getDosTrailer().checksum;
        else
            cksums[i-1] = 0;
    }

 have_block:
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (trns[i-1]) {
            trns[i-1]->setTemplock(cksums);
            trns[i-1]->updateTrailer();
        }
    }

    /* Write 'em out */
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (trns[i-1]) {
            Ptr<Stream> out = dir.openFile(format("player%d.trn", i), Stream::C_CREATE + Stream::C_WRITE + Stream::C_READ);
            console.write(LOG_INFO,
                          format(_("Writing %s turn file (%d command%!1{s%})..."),
                                 player_racenames.getAdjName(i),
                                 trns[i-1]->getNumCommands()));
            trns[i-1]->write(*out);

            // Write backup copies
            GBackupFileTemplate tpl;
            ProgressMonitor pm;
            tpl.setTurnNumber(trn.getTurnNumber());
            tpl.setPlayerNumber(i);
            tpl.setGameDirectory(dir.getName());

            out->seek(0);
            tpl.copyFile(opt_BackupTurn(), *out, pm);
            out->seek(0);
            tpl.copyFile(opt_MaketurnTarget(), *out, pm);
        }
    }
}
