/**
  *  \file int/if/objif.cc
  *  \brief Interpreter Game Interface: Common Object Methods
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/objif.h"
#include "int/if/ifutil.h"
#include "game/obj.h"
#include "int/interror.h"
#include "int/args.h"
#include "int/exec.h"

/** Constructor.
    \param obj Object to operate on
    \param call Function to call */
IntObjectProcedureValue::IntObjectProcedureValue(GObject& obj, GUniverse& univ, call_t call)
    : obj(obj), univ(univ), call_function(call)
{ }

/** Destructor. */
IntObjectProcedureValue::~IntObjectProcedureValue()
{ }

// IntCallableValue:
void
IntObjectProcedureValue::call(IntExecutionContext& exc, IntDataSegment& args, bool want_result)
{
    IntArgBlock a(args, 0, args.size());
    call_function(obj, univ, exc, a);
    if (want_result) {
        exc.pushNewValue(0);
    }
}

bool
IntObjectProcedureValue::isProcedureCall()
{
    return true;
}

int32_t
IntObjectProcedureValue::getDimension(int32_t)
{
    return 0;
}

IntContext*
IntObjectProcedureValue::makeFirstContext()
{
    return 0;
}

// IntValue:
string_t
IntObjectProcedureValue::toString(bool /*readable*/)
{
    return "#<obj-procedure>";
}

void
IntObjectProcedureValue::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

IntValue*
IntObjectProcedureValue::clone()
{
    return new IntObjectProcedureValue(obj, univ, call_function);
}

/******************************* Functions *******************************/

/** Implementation of the "Mark" command.
    Syntax: 'Mark [flag]' */
void
IFObjMark(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    /* @q Mark Optional flag:Bool (Planet Command, Ship Command, Ufo Command, Storm Command, Minefield Command)
       Mark object.
       Marks the current object.
       If the flag is specified as False, unmarks it instead.
       @since PCC2 1.99.9, PCC 1.0.5
       @diff This command is also available for ufos, ion storms, and minefields since PCC2 1.99.13.
       Older versions and PCC 1.x only allow it for ships and planets.
       @see Unmark */
    args.checkArgumentCount(0, 1);

    bool state = true;
    checkBoolArg(state, args.getNext());
    obj.setIsMarked(state);
}

/** Implementation of the "Unmark" command.
    Syntax: 'Unmark' */
void
IFObjUnmark(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    /* @q Unmark (Planet Command, Ship Command, Ufo Command, Storm Command, Minefield Command)
       Unmark object.
       Unmarks the current object.
       @since PCC2 1.99.9, PCC 1.0.5
       @diff This command is also available for ufos, ion storms, and minefields since PCC2 1.99.13.
       Older versions and PCC 1.x only allow it for ships and planets.
       @see Mark */
    args.checkArgumentCount(0);
    obj.setIsMarked(false);
}
