/**
  *  \file int/if/userif.cc
  *  \brief Interpreter: User Interface Binding
  *
  *  Unlike contexts (which are set when a script starts, and remain
  *  active for the complete script execution time), the user interface
  *  bindings vary during execution, and can be modified in all
  *  directions by scripts.
  *
  *  Scripts interact with the IntUserInterfaceBinding by posting
  *  commands and by querying and setting properties.
  *
  *  Every place that might execute scripts or otherwise refer to a
  *  IntUserInterfaceBinding must have an event loop that follows
  *  certain conventions:
  *  - if a command event has been posted, it must be processed.
  *    A possible processing could be to exit this loop, destroy the
  *    current binding, and hand the event to the parent binding.
  *  - otherwise, if there are runnable processes, they must run.
  *    These are the processes that have triggered the command or
  *    this very dialog, and want to continue.
  *  - only if that has been done, user interaction can occur.
  *  It is up to the dialog how to implement the user interaction.
  *  If it uses a regular UIWindow::run loop, postEvent must set
  *  aside the event and make sure that this loop exits for the event
  *  to be processed. In no case, postEvent must directly process
  *  the command.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/userif.h"
#include "int/if/ifutil.h"
#include "int/interror.h"

namespace {
    IntUserInterfaceRootBinding rootBinding;
    IntUserInterfaceBinding* currentBinding = &rootBinding;
}

/************************ IntUserInterfaceBinding ************************/

IntUserInterfaceBinding::IntUserInterfaceBinding()
{ }

IntUserInterfaceBinding::~IntUserInterfaceBinding()
{ }

IntUserInterfaceBinding&
IntUserInterfaceBinding::get()
{
    return *currentBinding;
}

/********************** IntUserInterfaceRootBinding **********************/

IntUserInterfaceRootBinding::IntUserInterfaceRootBinding()
{ }

IntUserInterfaceRootBinding::~IntUserInterfaceRootBinding()
{ }

IntValue*
IntUserInterfaceRootBinding::getProperty(IntUserInterfaceProperty /*p*/)
{
    return 0;
}

void
IntUserInterfaceRootBinding::setProperty(IntUserInterfaceProperty /*p*/, IntValue* /*v*/)
{
    throw IntError::notAssignable();
}

void
IntUserInterfaceRootBinding::postEvent(std::auto_ptr<IntUserInterfaceEvent> /*event*/)
{
    throw IntError::contextError();
}

void
IntUserInterfaceRootBinding::enumContexts(ptr_vector<IntContext>& /*ctx*/)
{ }

/********************** IntUserInterfaceChildBinding *********************/

IntUserInterfaceChildBinding::IntUserInterfaceChildBinding()
    : parent(get())
{
    currentBinding = this;
}

IntUserInterfaceChildBinding::~IntUserInterfaceChildBinding()
{
    if (currentBinding == this) {
        currentBinding = &parent;
    }
}
