/**
  *  \file io/dirs.h
  *  \brief Directory Handling
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_FILEIO_DIRS_H
#define PCC_V2_FILEIO_DIRS_H

#include <string>
#include "io/streams.h"
#include "cpluslib/smartptr.h"
#include "cpluslib/string.h"
#include "cpluslib/enum.h"

class Directory;

// DirectoryEntry
class DirectoryEntry : public Object {
 public:
    /** File position/size type. */
    typedef Stream::fpos_t fpos_t;

    /** Directory entry type. */
    typedef char FileType;

    /** Possible file types. */
    enum {
        FT_UNKNOWN    = '?',    ///< File type can't be detected
        FT_FILE       = '-',    ///< Regular file
        FT_DIRECTORY  = 'd'     ///< Directory
    };

 protected:
    /** Name of file. */
    const string_t name;

    /** Obtain information about this file. Must eventually call
        setInfo() on success; or omit this on failure. */
    virtual void getInfo() = 0;
    void setInfo(fpos_t size, FileType type);
    void clearInfo();
    DirectoryEntry(string_t name);

 public:
    virtual ~DirectoryEntry();

    /** Open this as a directory. Returns a Directory descendant object, or throws. */
    virtual Ptr<Directory> openDir() = 0;

    /** Open this as a file. Returns a Stream descendant object, or throws. */
    virtual Ptr<Stream> openFile(int mode) = 0;

    /** Open parent directory. */
    virtual Ptr<Directory> openParent() = 0;

    /** Erase file. The DirectoryEntry object continues to refer
        to the file name. Your implementation should call clearInfo(). */
    virtual void erase() = 0;

    /** Rename directory entry. The DirectoryEntry object continues to refer
        to the old name. Your implementation should call clearInfo(). */
    virtual void renameTo(string_t name) = 0;

    /** Create directory. Your implementation should call clearInfo()
        to forget previously-stored information. */
    virtual void createAsDirectory() = 0;

    /** Get name of this directory. (hmm, I wonder whether this has any use?) */
    string_t getName() const
        { return name; }
    virtual string_t getBaseName() const;

    FileType getFileType();
    fpos_t getSize();
 private:
    bool info_set;              ///< true iff file_size/file_type are set.
    fpos_t file_size;           ///< size of file in bytes. -1 on error.
    FileType file_type;         ///< type of file (FT_xxx).
};

// DummyDirectoryEntry
class DummyDirectoryEntry : public DirectoryEntry {
    Ptr<Directory> parent;
 protected:
    void getInfo();
 public:
    DummyDirectoryEntry(string_t name, Ptr<Directory> parent);
    Ptr<Directory> openDir();
    Ptr<Stream> openFile(int mode);
    Ptr<Directory> openParent();
    void erase();
    void renameTo(string_t);
    void createAsDirectory();
};

class Directory : public Object {
 public:
    virtual ~Directory();
    virtual Ptr<DirectoryEntry> getEntryByName(string_t name) = 0;
    virtual Ptr<Enumeration<Ptr<DirectoryEntry> > > getEntries() = 0;
    virtual Ptr<Directory> getParent() = 0;
    virtual string_t getName() = 0;
    virtual string_t getTitle();

    /* Convenience Functions */
    Ptr<Directory> openSubdir(string_t name);
    Ptr<Stream>    openFile(string_t name, int mode);
    Ptr<Stream>    openFileNT(string_t name, int mode);

    void erase(string_t name);
    bool eraseNT(string_t name);

    /* Static functions, platform specific (implemented in arch/) */
    static bool     isAbsoluteName(string_t name);
    static bool     isSeparatorChar(char c);
    static string_t makeFileName(string_t dir, string_t name);
    static string_t getCanonicalName(string_t dir);
    static string_t getAbsoluteName(string_t dir);
    static string_t getFileName(string_t dir);
    static string_t getDirName(string_t dir);
    static string_t getWorkingDirectory();
    static Ptr<Directory> openDir(string_t dir);
    static Ptr<Enumeration<Ptr<Directory> > > getRoots();
};

string_t appendFileNameExtension(string_t name, string_t ext, bool force);

#endif
